# Smileys Be Gone!
# A game with the goal to make smileys disappear
# Copyright (C) 2006 Neeraj Kumar and Robert Marianski

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import pygame
from pygame.locals import *
import pygame.mixer as mixer
from math import *
from gameutils import *
import os, os.path

class BaseSound(object):
	"""Base class for any sound or music"""

	SOUND, MUSIC = range(2)
	dirs = {SOUND: "sounds", MUSIC: "music"}
	snds = {}
	volume = {SOUND: 0.7, MUSIC: 0.5}

	def __init__(self, type, fnameGlob=None):
		"""initializes all the sounds given in the fname"""
		from glob import glob
		assert type in BaseSound.dirs
		self.type = type
		self.fnameGlob = fnameGlob
		snds = []
		self.numchannels = mixer.get_num_channels()
		path = os.path.join(BaseSound.dirs[type], fnameGlob)
		self.fnames = glob(path)
		self.fnames.sort()
		if fnameGlob not in self.__class__.snds.keys():
			if not fnameGlob: return
			#print "Reading files: ", self.fnames
			assert(len(self.fnames) > 0)
			# load all the sounds
			for fname in self.fnames:
				if type == BaseSound.SOUND:
					snd = mixer.Sound(fname)
					snds.append(snd)
			self.__class__.snds[fnameGlob] = snds
		self.setSound(0)

	def __len__(self):
		"""Returns the number of different sounds to choose from in this base sound"""
		return len(self.fnames)

	def setSound(self, num):
		"""Sets the soundfile at the given index to the active one."""
		if self.type == BaseSound.SOUND:
			self.snd = self.__class__.snds[self.fnameGlob][num]
		else:
			self.snd = mixer.music.load(self.fnames[num])
	
	def play(self, lrpos=0.5, repeat=0):
		"""Plays the currently selected sound. If repeat != 0, repeats until stop()-ed.
		The lrpos argument tells where (left-right) the sound is located. 0.0 means all the way left,
		1.0 means all the way right."""
		if self.type == BaseSound.SOUND:
			chan = mixer.find_channel()
			if not chan: return
			vals = [v*BaseSound.volume[BaseSound.SOUND] for v in [1.0-lrpos, lrpos]]
			chan.set_volume(*vals)
			if self.type == BaseSound.SOUND:
				if repeat:
					chan.play(self.snd, -1)
				else:
					chan.play(self.snd, 0)
		else:
			mixer.music.set_volume(BaseSound.volume[BaseSound.SOUND])
			if repeat:
				mixer.music.play(-1)
			else:
				mixer.music.play(0)

	def stop(self, fadeout=0):
		"""Stops this sound. If fadeout > 0, fades out over given number of ms"""
		if self.type == BaseSound.SOUND:
			self.snd.stop()
		else:
			if fadeout:
				mixer.music.fadeout(fadeout)
			else:
				mixer.music.stop()

class SoundEffect(BaseSound):
	"""A simple sound effect class that just plays when inited, and then dies"""
	SPRING, CARCRASH, MONEY, TRAIN = range(4)
	types = {SPRING: "boing*ogg", CARCRASH: "carcrash*ogg", MONEY: "chaching*ogg", TRAIN: "trainwhistle*ogg"}

	def __init__(self, type, x=None, screenwidth=600):
		"""Plays the sound of the given type. If an x-pos is given, then it places the sound in stereo"""
		assert type in SoundEffect.types
		BaseSound.__init__(self, BaseSound.SOUND, SoundEffect.types[type])
		self.setSound(randint(0, len(self)-1))
		pos = 0.5
		if x:
			pos = x/float(screenwidth)
			if pos < 0.0: pos = 0.0
			if pos > 1.0: pos = 1.0
		self.play(lrpos=pos)

class LevelMusic(BaseSound):
	"""A simple class to play music for a level"""
	MENU, GAME = range(2)
	types = {MENU: "menu_*ogg", GAME: "game_*ogg"}
	def __init__(self, type):
		assert type in LevelMusic.types
		BaseSound.__init__(self, BaseSound.MUSIC, LevelMusic.types[type])
		self.setSound(randint(0, len(self)-1))

	def updateVolume(self):
		"""Sets the volume from BaseSound.volume[MUSIC]"""
		mixer.music.set_volume(BaseSound.volume[BaseSound.MUSIC])
