# Smileys Be Gone!
# A game with the goal to make smileys disappear
# Copyright (C) 2006 Neeraj Kumar and Robert Marianski

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# the graphics classes for the game
import pygame
from pygame.locals import *
from pygame.sprite import *
from math import *
from gameutils import *
from gamestructs import *
from gameaudio import BaseSound
import operator as op
import os, os.path


# Some constants
w = 15
sizes = {TRAIN: 2*w, ROAD: 2*w, PED: w, BUILDING: 4*w}

class SpriteRect(Sprite):
	"""A simple class to create a sprite with only a rect for collision purposes"""
	def __init__(self, rect):
		Sprite.__init__(self)
		self.rect = rect
		self.image = pygame.Surface(rect.size)
		self.image.set_colorkey((0,0,0), RLEACCEL)
		self.image.fill((0,0,0))

class AnimatedSprite(Sprite):
	"""A sprite which can show an animation"""
	imagesDir = "images"
	ims = {}
	def __init__(self, fnameGlob=None, colorkey=None, centerpos=(0,0)):
		from glob import glob
		Sprite.__init__(self)
		self.fnameGlob = fnameGlob
		ims = []
		path = os.path.join(AnimatedSprite.imagesDir, fnameGlob)
		self.fnames = glob(path)
		self.fnames.sort()
		if fnameGlob not in self.__class__.ims.keys():
			if not fnameGlob: return
			#print "Reading files: ", self.fnames
			assert(len(self.fnames) > 0)
			# load all the images
			for fname in self.fnames:
				im = pygame.image.load(fname).convert()
				if colorkey: im.set_colorkey(colorkey)
				ims.append(im)
				self.__class__.ims[fnameGlob] = ims
		else:
			ims = self.__class__.ims[fnameGlob]
		self.image = None
		self.rect = ims[0].get_rect()
		self.rect.center = centerpos
		self.pos = centerpos		# convenience attribute
		self.angle = 0
		self.setImage(0)

	def __len__(self):
		"""Returns the number of frames in this animation"""
		return len(self.fnames)

	def getImage(self, num):
		"""Returns the image at the given number"""
		return self.__class__.ims[self.fnameGlob][num]

	def setImage(self, num):
		"""Sets the image at the given index to the active one. Rotates by current angle"""
		self.image = self.__class__.ims[self.fnameGlob][num]
		if self.angle != 0:
			self.image = pygame.transform.rotate(self.image, self.angle)
		# change the rect, keeping the center the same
		oldcenter = self.rect.center
		self.rect = self.image.get_rect()
		self.rect.center = oldcenter

class RandomSmiley(AnimatedSprite):
	"""A class that display a smiley at a random position and size for a random amount of time"""
	# the time between respawns of the smiley (in number of calls to update())
	respawnTimeInterval = [30, 300]

	# the increment for in/de-creasing transparency
	alphaIncr = 2

	# max alpha
	maxAlpha = 150

	def __init__(self, screensize, otherObjs):
		self.screensize = screensize
		self.otherObjs = otherObjs
		AnimatedSprite.__init__(self, fnameGlob="smileylarge*png")
		self.chooseNewConfig()

	def chooseNewConfig(self):
		"""Chooses a new configuration for our smiley"""
		found = 0
		maxiters = 100
		while not found:
			if maxiters < 0: break
			maxiters -= 1
			zoom = (random() * 0.3) + 0.1
			im = self.__class__.ims[self.fnameGlob][0]
			self.image = pygame.transform.rotozoom(im, 0, zoom).convert()
			self.image.set_colorkey((255, 255, 255))
			self.rect = self.image.get_rect()
			self.rect.center = [randint(0, self.screensize[0]), randint(0, self.screensize[1])]
			self.alpha = 0
			self.image.set_alpha(0)
			self.counter = randint(*RandomSmiley.respawnTimeInterval)
			found = 1
			for obj in self.otherObjs:
				colls = pygame.sprite.spritecollide(self, obj, 0)
				if colls:
					for c in colls:
						if c == self: continue
						found = 0
						break
		if found:
			self.needsReset = 0
		else:
			self.rect = (0, 0, 1, 1)
			self.counter = 300
			self.needsReset = 1

	def update(self):
		"""Updates our transparency"""
		if self.counter > 0:
			self.counter -= 1
			return
		if self.needsReset:
			self.chooseNewConfig()
			return
		# if we're less than 255, then we're increasing
		self.alpha += RandomSmiley.alphaIncr
		if self.alpha > 2*RandomSmiley.maxAlpha: 
			self.needsReset = 1
		self.image.set_alpha(RandomSmiley.maxAlpha - abs(RandomSmiley.maxAlpha-self.alpha))
	
class Button(AnimatedSprite):
	"""A simple button class that supports rollovers"""
	# the different states for a button
	NORMAL, ACTIVE = range(2)
	def __init__(self, fnameGlob, pos, obj=None, callback=None):
		"""Creates a new button at the given position"""
		AnimatedSprite.__init__(self, fnameGlob, centerpos=pos)
		self.setImage(0)
		self.callback = callback
		self.obj = obj
		self.state = 0
		self.setState(Button.NORMAL)

	def click(self):
		"""Clicks this button"""
		if self.callback: return self.callback()

	def setState(self, state):
		if state == self.state: return
		assert(state < len(self))
		self.setImage(state)
		self.state = state

class Light(AnimatedSprite):
	"""A car or pedestrian light"""
	# the time quanta for changing lights
	lightQuanta = 500

	def __init__(self, orientation, fnamePrefix, pos):
		assert(orientation in [HORIZONTAL, VERTICAL])
		self.orientation = orientation
		AnimatedSprite.__init__(self, fnameGlob="%s_%d_*.png" % (fnamePrefix, orientation), centerpos=pos)
		self.mintime = 0
		self.indices = []
		count = 0
		for i, t in enumerate(self.__class__.times):
			for x in range(count, count+t):
				self.indices.append(i)
			count += t
		s = sum(self.__class__.times)
		# the times must add up to an even number
		assert(s % 2 == 0)
		if orientation == VERTICAL:
			for i in range(s/2): self.indices.append(self.indices.pop(0))
		self.state = None
		self.setState(self.indices[0])
	
	def setState(self, state):
		"""Sets this light to the given state"""
		if self.state == state: return
		self.setImage(state)
		self.state = state

	def setQuanta(self, quanta, curtime):
		"""Sets the current quanta number"""
		self.setState(self.indices[quanta])

	def click(self, curtime):
		"""Click this light"""
		self.setState((self.state+1)%len(self))
		self.mintime = curtime+self.__class__.clicktime

	def resetBackToTimings(self, curtime):
		"""Resets this light back to normal state"""
		self.mintime = curtime
		
	def absMeth(self):
		abstract()

class CarLight(Light):
	"""A traffic light for cars"""
	# the timings for the car lights (green/yellow/red), in terms of lightQuanta
	times = [8, 2, 10]
	# the amount of time the light will stay the way a user clicked it
	clicktime = 10000
	GREEN, YELLOW, RED = range(3)
	def __init__(self, orientation, pos):
		Light.__init__(self, orientation, "carlight", pos)

class PedLight(Light):
	"""A traffic light for peds"""
	# the timings for the pedestrian lights (walk, stop), in terms of lightQuanta
	times = [6, 14]
	# the amount of time the light will stay the way a user clicked it
	clicktime = 10000000
	WALK, STOP = range(2)
	def __init__(self, orientation, pos):
		Light.__init__(self, orientation, "pedlight", pos)

class BaseMovingObject(AnimatedSprite):
	"""An object which keeps track of its velocity"""
	def __init__(self, fnameGlob, pos, curloc, colorkey=None, vel=[0.0, 0.0], target=None):
		AnimatedSprite.__init__(self, fnameGlob, colorkey=colorkey, centerpos=[int(p) for p in pos])
		self.pos = pos
		self.vel = vel
		self.curloc = curloc
		self.target = target
		self.curloc.addChild(self)

	def applyVelocity(self, timestep):
		"""Applies velocity over given timestep"""
		if timestep < 1: return
		# note that we must do our calculations in floating point, in case we only move slightly
		dpos = [v*timestep for v in self.vel]
		#if self.__class__ == Car and self.path: print "Applying vel: ", dpos, self.vel, self.pos
		self.pos = map(op.add, self.pos, dpos)
		#print "  Rect.center was: ", self.rect.center
		#print "  New center pos: ", self.pos
		# we now write out the new position to our rect
		self.rect.center = [int(p) for p in self.pos]
		#print "  New Rect.center: ", self.rect.center

	def move(self, timestep):
		"""Default move for moving objects is simply to apply their current velocities"""
		self.applyVelocity(timestep)

	def kill(self):
		"""Kills self, performing some other cleanup"""
		self.curloc.removeChild(self)
		AnimatedSprite.kill(self)

class Train(BaseMovingObject):
	"""A train which moves horizontally"""
	def __init__(self, node, trainspeed):
		"""Spawn a train starting at the given node"""

		# find the edge we're gonna be travelling on
		assert len(node.edges) == 1
		edge = node.edges[0]

		# figure out our orientation
		orientation = edge.orientation
		assert orientation in [HORIZONTAL, VERTICAL]

		# figure out the direction we're going in and set our target
		if node == edge.nodes[0]: direction = FORWARD
		else: direction = BACKWARD

		# get a location from the node
		pos = [float(i) for i in node.pos]
		# use our direction and the orientation to get in the right lane and to set our velocity
		pos[1-orientation] += direction*(sizes[TRAIN]/4+1)
		vel = [trainspeed*v*direction for v in unitVecs[orientation]]

		# call the super's initializer
		BaseMovingObject.__init__(self, 'train*.png', pos, edge, colorkey=(0, 0, 0), vel=vel)
		self.angle = vecToAngle(vel)
		self.setImage(0)
		# set our target
		self.target = edge.getOppositeNode(node)
		# figure out how far we are from the target
		self.lastDistance = distance(self.pos, self.target.pos)
	
	def update(self, curtime):
		d = distance(self.pos, self.target.pos)
		# if we're farther now from our target than we were last time, we must have overshot it. suicide
		if d > self.lastDistance:
			self.kill()
		self.lastDistance = d

class Car(BaseMovingObject):
	"""A car which moves horizontally or vertically"""
	ALIVE, DEAD = range(2)

	# the max car speed (pixels/ms)
	maxSpeed = 0.05

	# the maximum amount a car can accelerate or decelerate by (pixels/ms/ms)
	maxAccel = 0.003

	# the minimum distance cars will try to keep between each other (pixels)
	carMinDist = 30.0

	# the minimum distance cars will try to keep to a light (pixels)
	lightMinDist = 30.0

	# the maximum distance to a light that we care about (pixels)
	lightMaxDist = 120.0

	# the amount of time to show our dead body (ms)
	dyingTime = 5000

	def __init__(self, edge, percent, direction):
		"""Creates a car at the given percentage on an edge, going in the given direction (either"""
		self.orientation = edge.orientation
		assert self.orientation in [HORIZONTAL, VERTICAL]
		self.state = Car.ALIVE
		self.path = None
		self.oldloc = None
		assert direction in [FORWARD, BACKWARD]
		self.direction = direction
		# get a location from the edge
		pos = [float(i) for i in edge.getPositionAt(percent)]
		# use our direction and the orientation to get in the right lane
		if self.orientation == HORIZONTAL: pos[1] += direction*(sizes[ROAD]/4+1)
		else: pos[0] -= direction*(sizes[ROAD]/4+1)
		# set our target
		if direction == FORWARD: target = edge.nodes[1]
		else: target = edge.nodes[0]
		BaseMovingObject.__init__(self, 'car_*.png', pos, edge, colorkey=(255, 255, 255), target=target)
		# set our acceleration to 0
		self.accel = 0.0
		# choose a random car image
		self.carnum = randint(0, len(self)-1)
		self.setImageBasedOnOrientation()
		self.deathTime = 0
		self.killTime = 0

	def setImageBasedOnOrientation(self):
		"""Figures out our current orientation and sets the appropriate image"""
		if isinstance(self.curloc, Node):
			# if we're on a node, then it's the direction from our pos to the next element in the path
			if len(self.path) > 0:
				target = self.path[0]
			# if there's nothing left in the path, then we must be pointing in the right direction
			else: return
			dir = getDir(self.pos, target)
		else:
			# if we're on an edge, figure it out based on orientation and direction
			dir = [0.0, 0.0]
			dir[self.orientation] = self.direction
		self.angle = -((vecToAngle(dir) + 360) % 360)
		self.setImage(self.carnum)

	def getIntoLane(self):
		"""Simple function to make sure we're right in the middle of our lane"""
		# use our direction and the orientation to get in the right lane
		if isinstance(self.curloc, Node): return
		if self.orientation == HORIZONTAL: self.pos[1] = self.curloc.pos[1] + self.direction*(sizes[ROAD]/4+1)
		else: self.pos[0] = self.curloc.pos[0] - self.direction*(sizes[ROAD]/4+1)

	def adjustAccelForCars(self, timestep):
		"""Adjusts our acceleration to try to stay a certain distance behind the car in front of us"""
		self.aheadCar = None
		curspeed = getMag(self.vel)
		# find the closest car ahead of us
		aheads = []
		for c in self.curloc.children:
			# ignore ourselves, cars in the opposite lane, cars behind us
			if c == self: continue
			if c.direction != self.direction: continue
			if self.direction == FORWARD and c.pos < self.pos: continue
			if self.direction == BACKWARD and c.pos > self.pos: continue
			dist = distance(self.pos, c.pos)
			aheads.append((dist, c))
		if len(aheads) == 0: return
		aheads.sort()
		dist, self.aheadCar = aheads[0]
		if dist < Car.carMinDist:
			# if we're too close, slam on the brakes
			self.accel = min(max(-Car.maxAccel, -curspeed/timestep), self.accel)
			if fabs(curspeed) < 0.01: self.accel = 0
			return
		else:
			# otherwise, try to get within minDist at the next timestep
			aheadSpeed = getMag(self.aheadCar.vel)
			if curspeed < Car.maxSpeed:
				# we want to try to get to minDist behind the car at the next timestep	
				goalDist = dist + aheadSpeed*timestep - Car.carMinDist
				# find the speed we need to be at to get there at the next timestep
				goalSpeed = min(goalDist/timestep, Car.maxSpeed)
				# find the acceleration to apply to get to that speed
				self.accel = min((goalSpeed-curspeed)/timestep, Car.maxAccel)

	def adjustAccelForLight(self, timestep):
		"""Adjusts our acceleration to try to not run any lights"""
		# see if the light ahead of us is close enough
		e = self.curloc
		for n in range(2):
			if self.target == e.nodes[n]: light = e.lights[n]
		dist = distance(self.pos, light.rect.center)
		curspeed = getMag(self.vel)
		# we're too far away from the light to care
		if dist > Car.lightMaxDist: return
		# depending on the color of the light, do something
		if light.state == CarLight.RED:
			if dist < Car.lightMinDist:
				# we're too close to the light, slam on the brakes
				self.accel = min(max(-Car.maxAccel, -curspeed/timestep), self.accel)
				if fabs(curspeed) < 0.001: self.accel = 0
			else:
				# otherwise, slow down at a constant rate
				self.accel = max(-Car.maxAccel, -curspeed*curspeed*0.2/dist)
		if light.state == CarLight.YELLOW:
			# if the light's yellow, and we're very close to the light, gun it
			if dist < Car.lightMinDist:
				self.accel = (2*Car.maxSpeed - curspeed)/timestep
			
	def adjustAcceleration(self, timestep):
		"""Figure out our new acceleration, based on our surroundings"""
		if isinstance(self.curloc, Node):
			# if we're on a node, we won't use acceleration
			self.accel = 0.0
			return
		# otherwise, we must be on an edge. We need to check for cars, lights, and our curspeed
		# first try to go as fast as possible, then adjust down for cars, lights
		curspeed = getMag(self.vel)
		self.accel = max(min(Car.maxAccel, (Car.maxSpeed-curspeed)/timestep), -Car.maxAccel)
		self.adjustAccelForLight(timestep)
		self.adjustAccelForCars(timestep)
		self.path = None

	def applyAcceleration(self, timestep):
		"""Applies our stored acceleration to update the velocity"""
		if timestep < 1: return
		# if we're on a node, then use our path to figure out where to go
		if isinstance(self.curloc, Node):
			# based on our speed, figure out which index of the path we wanna go to
			curspeed = getMag(self.vel)
			if len(self.path) == 0: return
			indexnum = min(max(int(curspeed*timestep), 0), len(self.path)-1)
			destpos = self.path[indexnum]
			# calculate the vector to this location
			vec = getDir(self.pos, destpos)
			# our velocity is then whatever's needed to get traverse this distance in this timestep
			self.vel = magToVec(curspeed, vec)
			# kill everything in the path through the indexnum
			self.path = self.path[indexnum+1:]
			self.setImageBasedOnOrientation()
			return
		# otherwise, we just continue on our edge, either slower or faster
		accel = self.direction * self.accel
		# convert the scalar acceleration to a vector one
		if self.orientation == HORIZONTAL:
			accel = [accel, 0.0]
			self.vel[1] = 0.0
		else:
			accel = [0.0, accel]
			self.vel[0] = 0.0
		# now add this to the velocity
		dvel = [a*timestep for a in accel]
		self.vel = map(op.add, self.vel, dvel)

	def move(self, timestep):
		"""Moves this car over the course of the given timestep"""
		if self.state == Car.DEAD: return
		self.applyAcceleration(timestep)
		self.applyVelocity(timestep)

	def update(self, curtime):
		# based on our current position, figure out what node or edge we should be on
		# if we're no longer within our edge or node, then we should be in our target
		if self.state == Car.DEAD:
			# if the car has been dead for enough time, remove it from the screen
			if curtime > self.killTime:
				self.kill()
			return
		if not self.curloc.getRect().collidepoint(*self.pos):
			#TODO this fails every once in a while
			#assert self.target.getRect().collidepoint(*self.pos)
			# choose a new target
			self.oldloc = self.curloc
			self.oldloc.removeChild(self)
			self.curloc = self.target
			self.curloc.addChild(self)
			if isinstance(self.curloc, Node):
				# we went from edge to node so choose a new edge to go to
				#print "Going from edge to node"
				if len(self.curloc.edges) == 1:
					# if this node has only one edge, then it must be a spawn point
					self.kill()
					return
				# otherwise, pick a new edge to go to
				choices = sample(self.curloc.edges, len(self.curloc.edges))
				self.target = choices[0]
				if self.target == self.oldloc: self.target = choices[1]
				# calculate and save our path
				self.path = self.curloc.getPath(self.oldloc, self.target, self.pos)
			else:
				# we went from node to edge, so choose the opposite node to go to
				#print "Going from node to edge"
				nodes = self.curloc.nodes
				if nodes[0] == self.oldloc:
					self.target = nodes[1]
					self.direction = FORWARD
				else:
					self.target = nodes[0]
					self.direction = BACKWARD
				self.orientation = self.curloc.orientation
				# make sure we're right in the middle of the lane
				self.getIntoLane()
				self.setImageBasedOnOrientation()
			assert self.oldloc != self.target

	def kill(self):
		"""Kills ourself"""
		self.angle = 0
		self.state = Car.DEAD
		if self.deathTime > self.killTime:
			self.killTime = self.deathTime + Car.dyingTime
		else:
			BaseMovingObject.kill(self)

class Ped(BaseMovingObject):
	WALKING, FLYING, DEAD = range(3)
	walkingSpeed = 0.01
	drag = 0.001
	# the amount of time to show our dead body (ms)
	dyingTime = 3000
	"""A pedestrian."""
	def __init__(self, edge, percent, direction, pedgroups, pednodes, pededges):
		"""Spawn a walking pedestrian on this edge"""
		assert direction in [FORWARD, BACKWARD]
		self.orientation = edge.orientation
		self.state = Ped.WALKING
		self.direction = direction
		self.pedgroups = pedgroups
		self.pednodes = pednodes
		self.pededges = pededges
		self.numBounces = 0
		# get a location from the edge
		pos = [float(i) for i in edge.getPositionAt(percent)]
		# set our target
		if direction == FORWARD: target = edge.nodes[1]
		else: target = edge.nodes[0]
		# get a vector to the target and set our velocity in that direction
		dirToTarget = getDir(pos, target.pos)
		vel = magToVec(Ped.walkingSpeed, dirToTarget)
		BaseMovingObject.__init__(self, "smiley*.png", pos, edge, colorkey=(255,255,255), vel=vel, target=target)
		# this will point to a light that we're waiting on
		self.waiting = None
		self.oldscores = (0.0, 0.0)
		# make sure we don't get bounced by the same spring multiple times
		self.lastSpring = None
		self.onSidewalk = 1
		self.killTime = 0
		self.deathTime = 0

	def move(self, timestep):
		"""Moves the ped around, depending on his current state"""
		if self.state == Ped.WALKING:
			if not self.onSidewalk:
				# if we're not on sidewalk, just keep walking until we hit it
				self.applyVelocity(timestep)
				return
			curloc = self.curloc
			# if we're on a node, then go unless we're waiting for a light
			if isinstance(curloc, Node):
				if self.waiting and self.waiting.state == PedLight.STOP:
					#print "WAITING FOR A RED LIGHT at ", self.pos
					self.vel = [0.0] * 2
			else:
				# otherwise we must be on an edge
				# first see if there's some money on our own edge
				objs = []
				for obj in [c for c in curloc.children if isinstance(c, ForceObject)]:
					# divide the score by its distance from us to get a better approximation
					dist = distance(obj.pos, self.pos)
					objs.append((obj.score/dist, obj.pos))
				pickedDir = 0
				if len(objs) > 0:
					# sort the objects and go for the first one, unless it's at negative score
					objs.sort(reverse=1)
					for score, pos in objs:
						if score > 0:
							orient = curloc.orientation
							if pos[orient] > self.pos[orient]: self.setTarget(curloc.nodes[1])
							else: self.setTarget(curloc.nodes[0])
							assert (self.target.pos[orient]-pos[orient] > 0) == (pos[orient]-self.pos[orient] > 0)
							pickedDir = 1
							break
				if not pickedDir:
					# there was nothing on our edge, so evaluate both ends to see if we want to change dirs
					s0 = curloc.nodes[0].score
					s1 = curloc.nodes[1].score
					if self.oldscores != (s0, s1):
						#print id(self), "saw that something changed, new probs are %f, %f" %(s0, s1)
						# choose using the scores as probabilities
						if s0 > s1: self.setTarget(curloc.nodes[0])
						elif s1 > s0: self.setTarget(curloc.nodes[1])
						else: self.setTarget(choice(curloc.nodes))
					self.oldscores = (s0, s1)
		elif self.state == Ped.FLYING:
			# slow them down because of wind drag
			drag = magToVec(min(Ped.drag, getMag(self.vel)), self.vel)
			self.vel = [v-d for v,d in zip(self.vel, drag)]
			if getMag(self.vel) < Ped.walkingSpeed:
				# if they slowed down enough, then figure out where they are and get them back to safety, asap
				self.recoverFromFlying()
			# rotate them while flying
			self.angle += 160.0*getMag(self.vel)
			self.setImage(self.state)
		# apply the new velocity
		self.applyVelocity(timestep)

	def chooseNewGoal(self, lastLoc=None):
		"""Choose a new goal for ourselves, depending current position, scores, etc.
		If lastLoc is given, then we try to avoid going back there"""
		assert self.onSidewalk
		assert self.state == Ped.WALKING
		self.waiting = None
		if isinstance(self.curloc, Node):
			# if we're on a node, then we need to choose an edge
			# take into account the scores of each node
			objs = []
			for e in self.curloc.edges:
				n = e.getOppositeNode(self.curloc)
				dist = distance(n.pos, self.pos)
				s = n.score/dist
				# if this was our last location, reduce the score for it
				if e == lastLoc: s /= 3.0
				# if this edge contains money on it, emphasize it
				for obj in [c for c in e.children if isinstance(c, ForceObject)]:
					if obj.score > 0:
						s *= obj.score
						break
				objs.append((s, e))
			if len(objs) > 0:
				# if we have some options, then pick the best one
				objs.sort(reverse=1)
				self.setTarget(objs[0][1])
			assert self.target in self.curloc.edges

			# if there are any ped lights, then see which one we need to obey (if any)
			if len(self.target.lights) > 0:
				for n in range(2):
					if self.curloc == self.target.nodes[n]: self.waiting = self.target.lights[n]
		else:
			if self.oldloc in self.curloc.nodes:
				# if we came onto this edge from one of it's nodes, choose the opposite one
				self.target = self.curloc.getOppositeNode(self.oldloc)
			else:
				# we came from somewhere else (i.e. from flying), so choose one probabilistically
				s0 = self.curloc.nodes[0].score
				s1 = self.curloc.nodes[1].score
				if s0 > s1: self.setTarget(self.curloc.nodes[0])
				elif s1 > s0: self.setTarget(self.curloc.nodes[1])
				else: self.setTarget(choice(self.curloc.nodes))
		# set our velocity to be in the direction of the new target
		dirToTarget = getDir(self.pos, self.target.pos)
		self.vel = magToVec(Ped.walkingSpeed, dirToTarget)

	def recoverFromFlying(self):
		"""If we survived our flying experience, then start walking around again.
		The first priority is to get back on sidewalk"""
		self.setState(Ped.WALKING)
		self.lastSpring = None
		self.oldloc = None
		self.numBounces = 0
		self.angle = 0
		# first see if we're on a sidewalk
		sidewalks = combinedSeq(self.pednodes, self.pededges)
		for s in sidewalks:
			if s.getRect().collidepoint(self.pos):
				# we're on this element, so set that as our position and choose our goal
				self.curloc = s
				self.onSidewalk = 1
				self.chooseNewGoal()
				#print id(self), "successfully landed on sidewalk after flying!"
				return
		# otherwise, we need to find the closest one and head in that direction
		self.onSidewalk = 0
		locs = []
		for s in sidewalks:
			# for each possible location, get its distance and score
			dist = distance(self.pos, s.pos)
			#try: score = s.score/dist
			#except Exception: score = 1.0/dist
			score = 1.0/dist
			locs.append((score, s))
		# now sort through our list of scores and choose the best one as the target
		locs.sort(reverse=1)
		self.setTarget(locs[0][1])

	def setState(self, state):
		"""Sets our state to the given one"""
		# remove ourselves from all previous groups
		assert state in [Ped.WALKING, Ped.FLYING, Ped.DEAD]
		walking, flying, dead = self.pedgroups
		self.remove(walking, flying, dead)
		self.state = state
		if self.state == Ped.WALKING:
			self.add(walking)
		elif self.state == Ped.FLYING:
			#print id(self), "just started flying at ", self.pos
			self.add(flying)
		elif self.state == Ped.DEAD:
			self.add(dead)
		self.setImage(self.state)

	def setTarget(self, target):
		"""Sets our new target and updates our velocity accordingly"""
		# if we're on sidewalk, the only valid target is an adjacent sidewalk
		if self.onSidewalk:
			if isinstance(self.curloc, Node): assert target in self.curloc.edges
			elif isinstance(self.curloc, Edge): assert target in self.curloc.nodes
		self.target = target
		dirToTarget = getDir(self.pos, self.target.pos)
		self.vel = magToVec(Ped.walkingSpeed, dirToTarget)

	def update(self, curtime):
		"""Based on our current position, figure out what node or edge we should be on"""
		# ignore flying peds
		if self.state == Ped.FLYING: return
		if self.state == Ped.DEAD:
			# if the ped has been dead for enough time, remove him from the screen
			if curtime > self.killTime:
				self.kill()
			return
		if not self.onSidewalk:
			# if we're trying to get back onto a sidewalk, then see if we've made it
			if self.target.getRect().collidepoint(*self.pos):
				# we've made it back, hallelujah!
				#print id(self), "just made it back to the sidewalk!"
				self.onSidewalk = 1
				self.curloc = self.target
				self.oldloc = self.target
				self.chooseNewGoal()
			return
			
		if not self.curloc.getRect().collidepoint(*self.pos):
			# if we're no longer within our edge or node, then we should be in our target
			hit = self.target.getRect().collidepoint(*self.pos)
			if not hit:
				#print "Should have been in target: ", self.target.getRect(), self.target, self.pos
				#assert hit
				pass
			# choose a new target
			self.oldloc = self.curloc
			self.oldloc.removeChild(self)
			self.curloc = self.target
			self.curloc.addChild(self)
			self.chooseNewGoal(self.oldloc)

	def kill(self):
		"""Kills ourselves, showing a death image"""
		# set ourselves to the dead state
		self.angle = 0
		self.setState(Ped.DEAD)
		if self.deathTime > self.killTime:
			self.killTime = self.deathTime + Ped.dyingTime
		else:
			BaseMovingObject.kill(self)

class PlaceableObject(AnimatedSprite):
	"""Any object which can be placed on the map"""
	def __init__(self, fnameGlob, pos, parent, incrfunc):
		AnimatedSprite.__init__(self, fnameGlob=fnameGlob, centerpos=pos, colorkey=(255, 255, 255))
		self.parent = parent
		self.incrfunc = incrfunc
		self.incrfunc(-1)

	def updateObject(self, curtime): 
		"""Updates the object as necessary"""
		abstract()

	def kill(self):
		"""Kills ourselves"""
		self.incrfunc(1)
		AnimatedSprite.kill(self)

class ForceObject(PlaceableObject):
	"""Objects which exert a force on people (either attractive or repulsive)"""
	MONEY, DOGSHIT = range(2)
	# The maximum distance out a forceobject affects stuff (pixels)
	maxDistance = 200.0
	types = {MONEY: ["money.png", 1000.0], DOGSHIT: ["dogshit.png", -1000.0]}
	def __init__(self, type, pos, parent, incrfunc):
		PlaceableObject.__init__(self, ForceObject.types[type][0], pos, parent, incrfunc)
		self.type = type
		self.pos = pos
		self.score = ForceObject.types[self.type][1]
		self.neighbors = []
		self.notifyParent()

	def notifyParent(self):
		"""notify our parent that we just got added and propagate the scores"""
		self.parent.addChild(self)
		neighbors = []
		# figure out all the neighboring nodes we wanna inform about ourselves
		if isinstance(self.parent, Node):
			# We placed this on a node. tell all the neighboring nodes that we have something
			neighbors.append(self.parent)
			for e in self.parent.edges:
				for n in e.nodes:
					if n != self.parent: neighbors.append(n)
		else:
			# We placed this on an edge. Tell the neighboring nodes that we have something,
			# and one more level beyond that
			neighbors.extend(self.parent.nodes[:])
			for nodes in neighbors:
				for e in nodes.edges:
					if e == self.parent: continue
					for n in e.nodes:
						if n in neighbors: continue
						neighbors.append(n)
		# save this list of neighbors and apply scores
		self.neighbors = neighbors
		self.changeScores(1.0)

	def changeScores(self, factor=1.0):
		"""Changes scores on our neighbors, by the given factor"""
		# add a score to each neighbor, which falls off as dist^2
		for n in self.neighbors:
			dist = distance(n.pos, self.pos)
			if dist > ForceObject.maxDistance: continue
			if (dist*dist) == 0: continue
			n.score += factor * self.score/(dist*dist)

	def kill(self):
		"""Kills the object, and notifies our parent that we're gone"""
		self.parent.removeChild(self)
		PlaceableObject.kill(self)
		# go through all our neighbors and subtract away our score
		self.changeScores(-1.0)
		self.neighbors = []

class Spring(PlaceableObject):
	"""A spring which flings people"""
	maxStrength = 0.4
	scaleFactor = 0.00042
	maxBounces = 10
	multiplier = 1.2

	READY, BUSY = range(2)
	interval = 20

	def __init__(self, pos, dir, parent, incrfunc):
		# based on the direction, choose a picture
		#angle = int(angle+360) % 360
		fnameglob = "spring_*.png"
		PlaceableObject.__init__(self, fnameglob, pos, parent, incrfunc)
		speed = min(Spring.scaleFactor*getMag(dir), Spring.maxStrength)
		self.vel = magToVec(speed, dir)
		self.bouncesleft = Spring.maxBounces
		self.angle = -((vecToAngle(self.vel) + 360) % 360)
		self.setImage(0)
		self.state = Spring.READY
		self.bouncedAt = 0

	def applyToPerson(self, p, curtime):
		"""Applies this spring to the given person. If it was actually applied, returns 1"""
		if self.state == Spring.BUSY: return 0
		if p.state == Ped.DEAD: return 0
		if self.bouncesleft <= 0: return 0
		if p.lastSpring == self: return 0
		self.state = Spring.BUSY
		self.bouncedAt = curtime
		p.setState(Ped.FLYING)
		p.lastSpring = self
		p.numBounces += 1
		# add the magnitude of their velocity to the spring direction
		curM = getMag(p.vel)
		springM = getMag(self.vel)
		springM *= Spring.multiplier * (p.numBounces)
		p.vel = magToVec(curM+springM, self.vel)
		self.bouncesleft -= 1
		if self.bouncesleft <= 0: self.kill()
		return 1

	def update(self, curtime):
		# if the spring was clicked, then change our image
		if self.state == Spring.BUSY:
			dif = curtime-self.bouncedAt
			num = dif/Spring.interval
			if num >= len(self):
				# we're done with the animation, so now actually bounce the person
				self.setImage(0)
				self.state = Spring.READY
			else:
				self.setImage(num)

class TrainSpawn(AnimatedSprite):
	"""A spawn point for a train. Note that this is the visual representation which can be clicked.
	The actual location where the train is spawned is off-screen, at the node"""

	# the range of intervals (in ms) for trains to spawn
	interval = (5000, 30000)

	# the speed of trains
	trainspeed = 0.1

	def __init__(self, node, screensize):
		self.node = node
		self.nextspawntime = randint(0, TrainSpawn.interval[1])
		self.mintime = 0
		
		assert len(node.edges) == 1
		edge = node.edges[0]
		orientation = edge.orientation
		# figure out where to place this spawn point based on orientation and edge position
		if node == edge.nodes[0]:
			# this is the top/left spawn point
			pos = [max(i, 0) for i in node.pos]
		elif node == edge.nodes[1]:
			# this is the bottom/right spawn point
			pos = [min(i, s) for i, s in zip(node.pos, screensize)]
		AnimatedSprite.__init__(self, fnameGlob="trainspawn_%d*.png" % (orientation), centerpos=pos)

	def spawn(self, curtime, clicked=0):
		# if this trainspawn is not ready to spawn a new one, continue
		#if clicked: print "Clicked and mintime is %d, curtime is %d" % (self.mintime, curtime)
		if not clicked and curtime < self.nextspawntime: return None
		if curtime < self.mintime: return None
		# set the next time the train will be spawned
		self.nextspawntime = curtime + randint(*TrainSpawn.interval)
		self.mintime = curtime + TrainSpawn.interval[0]
		# spawn the train
		t = Train(self.node, TrainSpawn.trainspeed)
		return t

class CarSpawn:
	"""Spawn point for a car. This corresponds to the node's location, unlike the trainspawns"""
	carOffset = 0.1

	def __init__(self, node):
		self.node = node

	def spawn(self, cars):
		"""Spawn a new car here if possible"""
		# keep choosing an edge until all edges have been exhausted
		choices = sample(self.node.edges, len(self.node.edges))
		for e in choices:
			# figure out which end we're at
			if self.node == e.nodes[0]:
				cur = 0
				dir = FORWARD
				loc = CarSpawn.carOffset
			else:
				cur = 1
				dir = BACKWARD
				loc = 1.0-CarSpawn.carOffset
			# create a car here and then see if there's anything already there
			c = Car(e, loc, dir)
			if pygame.sprite.spritecollideany(c, cars):
				# it collided with something--kill it and continue
				c.kill()
			else:
				# it didn't collide--add it to the list of cars and we're done
				cars.add(c)

class Flames(AnimatedSprite):
	"""A class for showing animated flames"""
	def __init__(self, pos, curtime, repeat=0, interval=80, callback=None):
		"""Creates the flames and starts them going"""
		AnimatedSprite.__init__(self, fnameGlob="flames*png", colorkey=(0, 0, 0), centerpos=pos)
		self.starttime = curtime
		self.interval = interval
		self.repeat = repeat
		self.callback = callback

	def update(self, curtime):
		"""Updates the flames"""
		# figure out what frame number we're at
		dif = curtime - self.starttime
		framenum = dif/self.interval
		# if we're past the end of the animation, figure out what to do
		if framenum >= len(self):
			# if we have repeat turned on, then just mod with number of frames
			if self.repeat:
				framenum = framenum % len(self)
			else:
				# otherwise kill
				self.kill()
				return
		self.setImage(framenum)
		self.rect.bottom = int(self.pos[1])+2

	def kill(self):
		"""Kills ourself and calls our callback, if we have one"""
		AnimatedSprite.kill(self)
		if self.callback: self.callback()

class TextButton(pygame.sprite.Sprite):
	"""Sprite text with no image"""
	pygame.font.init()	# prevents error on startup
	def __init__(self, text, onclick=lambda:None, color=(255,255,255), font=pygame.font.Font("gamefont.ttf", 20), activecolor=(255,0,0), **pw):
		"""Create a new text button. The pw kw's are the position keywords passed when creating the rect"""
		pygame.sprite.Sprite.__init__(self)
		self.font = font
		self.click = onclick
		self.color = color
		self.activecolor = activecolor
		self.pw = pw
		self._drawtext(text)
	def _drawtext(self, text, color=None):
		"""helper function to draw text and avoid duplication"""
		if color is None: color = self.color
		self.text = text
		self.image = self.font.render(self.text, 1, color)
		self.rect = self.image.get_rect(**self.pw)
	def setText(self, text, **pw):
		"""call if you want to change the text at some point"""
		if pw: self.pw = pw
		self._drawtext(text)
	def setState(self, state):
		"""change the color on the image"""
		if state == Button.ACTIVE:
			self._drawtext(self.text, self.activecolor)
		else:
			self._drawtext(self.text)

class IncrementingText(TextButton):
	"""responsible for handling the incrementing decrementing states"""
	def __init__(self, textstates, offset=50, onclick=lambda: None, color=(255,255,255), inccolor=None, font=pygame.font.Font("gamefont.ttf", 20), **pw):
		"""create the text groups and initialize the state"""
		if inccolor is None: inccolor = color
		self.states = textstates
		self.nstates = len(textstates)
		self.i = 0
		TextButton.__init__(self, textstates[0], color=color, font=font, **pw)
		# the offset to use for the buttons
		# now create the 2 increment/decrement buttons
		x,y = self.rect.center
		self.incbutton = TextButton(">", self.incclick, inccolor, font, centerx=x+offset, centery=y)
		self.decbutton = TextButton("<", self.decclick, inccolor, font, centerx=x-offset, centery=y)
		self.arrow = 1
	def incclick(self):
		"""increment click"""
		self.i += 1
		self._drawtext(self.states[self.i % self.nstates])
	def decclick(self):
		"""decrement click"""
		self.i -= 1
		if self.i < 0: self.i = self.nstates-1
		self._drawtext(self.states[self.i % self.nstates])
	def _adjustRects(self):
		self.incbutton.rect.centerx = self.rect.right+self.offset
		self.decbutton.rect.centerx = self.rect.left-self.offset
	def getClickableSprites(self):
		"""return the 2 sprites (< and >) that can be clicked on"""
		return (self, self.decbutton, self.incbutton)

class VolumeControl(AnimatedSprite):
	"""control the volume on the music"""
	# map from sound type to image glob
	globs = {BaseSound.SOUND: 'soundbutton_*.png',
	         BaseSound.MUSIC: 'musicbutton_*.png'}
	def __init__(self, soundtype, centerpos, notifyfn=lambda: None):
		"""initialize the music control sprite at the specified location"""
		assert soundtype in (BaseSound.SOUND, BaseSound.MUSIC)
		AnimatedSprite.__init__(self, fnameGlob=self.__class__.globs[soundtype], colorkey=(255,255,255), centerpos=centerpos)
		# global location of gameaudio volumes
		self.dictToModify = BaseSound.volume
		# refer to the appropriate sound intensity based on type
		self.soundtype = soundtype
		# current image number
		self.curimnum = len(self)-1
		self.setImage(self.curimnum)
		self.notifyfn = notifyfn
	def click(self):
		"""set the next sound intensity"""
		n = len(self) - 1
		if n == 0: n = 1
		self.curimnum += 1
		if self.curimnum >= len(self): self.curimnum = 0
		self.dictToModify[self.soundtype] = 0.8 * self.curimnum / n
		self.setImage(self.curimnum)
		self.notifyfn()

