# Smileys Be Gone!
# A game with the goal to make smileys disappear
# Copyright (C) 2006 Neeraj Kumar and Robert Marianski

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gameutils import *
from pygame import Rect
from random import *
import operator as op

# The different types of elements there are
TRAIN, ROAD, PED, BUILDING = range(4)

#TODO Note that right now we're saving pointers to nodes and edges directly, rather than storing
#     their ids. This might cause problems in the future, but it's ok for now, as it greatly simplifies
#     the code.

class GraphObject(object):
	def __init__(self, id, type, centerpos, size):
		self.children = []
		assert type in range(4)
		self.id = id
		self.type = type
		self.pos = centerpos
		self.size = size

	def getRect(self):
		"""Returns the bounding rect of this object"""
		abstract()
	
	def addChild(self, child):
		"""Adds a child as our child"""
		if child not in self.children:
			self.children.append(child)

	def removeChild(self, child):
		"""Removes a given child from our list of children"""
		if child in self.children:
			self.children.remove(child)

class Edge(GraphObject):
	def __init__(self, id, nodes):
		assert len(nodes) == 2
		p0, p1 = nodes[0].pos, nodes[1].pos
		type = nodes[0].type
		assert(nodes[0].type == nodes[1].type)
		centerpos = avgTuples([p0, p1])
		length = distance(p0, p1) - nodes[0].size
		if p0[0] == p1[0]: self.orientation = VERTICAL
		elif p0[1] == p1[1]: self.orientation = HORIZONTAL
		else: raise ValueError("Edges must be horizontally or vertically aligned")

		GraphObject.__init__(self, id, type, centerpos, length)
		self.nodes = list(nodes)
		if nodes[0].pos > nodes[1].pos: self.nodes.reverse()
		for n in self.nodes: n.addEdge(self)

	def __del__(self):
		"""Removes this edge"""
		for n in self.nodes: n.remEdge(self)

	def __contains__(self, node):
		"""Returns True if the given node is one of the nodes we have, False otherwise"""
		return node in self.nodes

	def getRect(self):
		"""Returns the bounding rect of this edge"""
		nodesize = self.nodes[0].size
		if self.orientation == HORIZONTAL:
			return Rect(self.pos[0]-self.size/2, self.pos[1]-nodesize/2, self.size, nodesize)
		elif self.orientation == VERTICAL:
			return Rect(self.pos[0]-nodesize/2, self.pos[1]-self.size/2, nodesize, self.size)
		else: raise ValueError("Edges must be horizontally or vertically aligned")

	def getPositionAt(self, percent):
		"""Returns a position on this edge a given percentage of the distance from its first node"""
		# convert the percentage to a distance
		dist = int(percent*self.size) - self.size/2
		if self.orientation == HORIZONTAL: return [self.pos[0]+dist, self.pos[1]]
		else: return [self.pos[0], self.pos[1]+dist]

	def getOppositeNode(self, node):
		"""Returns the node that's NOT the one given. Asserts if the one given is not in this edge"""
		assert node in self.nodes
		for n in self.nodes:
			if n != node: return n

class Node(GraphObject):
	def __init__(self, id, type, centerpos, size, isSpawn=0):
		GraphObject.__init__(self, id, type, centerpos, size)
		self.edges = []
		self.isSpawn = isSpawn

	def getRect(self):
		"""Returns the bounding rect of this object"""
		return Rect(self.pos[0]-self.size/2, self.pos[1]-self.size/2, self.size, self.size)

	def addEdge(self, e):
		"""Adds the given edge to our list of edges"""
		assert len(self.edges) < 4
		assert e.type == self.type
		self.edges.append(e)

	def remEdge(self, e):
		"""Removes the given edge"""
		self.edges.remove(e)
	
	def __del__(self):
		"""Removes this node, and all edges connected to it"""
		for e in self.edges: del e

	def __contains__(self, edge):
		"""Returns True if the given edge is one of the edges we have, False otherwise"""
		return edge in self.edges

	def getRelAngle(self, edge1, edge2):
		"""Returns the relative angle going from edge1 to edge2"""
		assert edge1 in self.edges
		assert edge2 in self.edges
		if edge1 == edge2: return 0
		dirs = [getDir(self.pos, e.pos) for e in [edge1, edge2]]
		angles = [vecToAngle(d) for d in dirs]
		return angles[1]-angles[0]

	def getPath(self, e1, e2, curpos):
		"""Returns a path to go from e1 to e2"""
		angle = self.getRelAngle(e1, e2)
		ang = int(((angle+360) % 360) / 90)
		path = []
		if ang == 0:
			# turning around
			pass
		elif ang == 2:
			# going straight
			# get a unit vector in the direction they're going
			dir = getDir(e1.pos, e2.pos)
			uvec = magToVec(1.0, dir)
			for i in range(self.size):
				curpos = map(op.add, curpos, uvec)
				path.append(curpos)
		elif ang == 1 or ang == 3:
			# turning left(1) or right(3)
			# get the corner around which the car is turning
			dir2 = getDir(self.pos, e2.pos)
			dir1 = getDir(self.pos, e1.pos)
			# these are the border points in each direction
			b1 = magToVec(self.size/2, dir1)
			b2 = magToVec(self.size/2, dir2)
			center = map(op.add, self.pos, b1)
			center = map(op.add, center, b2)
			# get the distance to this point (center of the circle which will be our path)
			radius = distance(center, curpos)
			# get the starting and ending angle
			startingangle = radians(vecToAngle(getDir(center, map(op.add, b1, self.pos))))
			# figure out how many radians to go to get unit-length steps
			step = 1/radius
			if ang == 1: step *= -1.0
			#print "radius, angdist, arclen, step: ", radius, angulardist, arclen, step
			# now generate all the points
			a = startingangle
			for i in range(int(radius*pi/2.0)):
				a += step
				pos = (radius*cos(a), radius*sin(a))
				pos = map(op.add, center, pos)
				path.append(pos)
			#print startingangle, ang, step, len(path)#, path
		#raw_input('asodifj')
		return path


class RoadNode(Node):
	def __init__(self):
		Node.__init__(self)
		self.pednodes = []
		self.type = ROAD

class PedNode(Node):
	def __init__(self):
		Node.__init__(self)
		self.roadnode = None
		self.type = PED

class Goal(object):
	"""A simple class encapsulating the idea of a goal"""
	def __init__(self):
		"""Basic init does nothing"""
		pass

	def passedGoal(self, *args):
		"""Returns True if we passed our goal. False otherwise"""
		abstract()

class DeathGoal(Goal):
	"""A goal where you need a certain number of deaths"""
	def __init__(self, deathsvar, minNeeded, minBounces=0, goalstr=None):
		"""Constructor which needs the list of deaths ('deathsvar'), 
		the index of the one you care about, and the minimum number of deaths needed"""
		self.deaths = deathsvar
		self.minnum = minNeeded
		self.minBounces = minBounces
		self.goalstr = goalstr
		self.passedGoal()

	def passedGoal(self):
		"""Checks to see if we've met our goal, and returns True if we have."""
		# go through the dictionary of deaths and add all those that have the minimum number of bounces
		self.num = 0
		for k, v in self.deaths.iteritems():
			if k < self.minBounces: continue
			self.num += v
		# if we have enough, then return true
		if self.num >= self.minnum: return True
		return False

	def __str__(self):
		if not self.goalstr:
			self.goalstr = "Death goal: %d/%d killed"
		s = self.goalstr % (self.num, self.minnum)
		return s
