# Smileys Be Gone!
# A game with the goal to make smileys disappear
# Copyright (C) 2006 Neeraj Kumar and Robert Marianski

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from math import *
from pygame import Rect
from random import *
import operator as op
from itertools import imap

# global enum for horizontal and vertical elements
HORIZONTAL, VERTICAL = range(2)
unitVecs = {HORIZONTAL: [1.0, 0.0], VERTICAL: [0.0, 1.0]}

# global enum for directions
FORWARD, BACKWARD = (1, -1)

def print_license():
	"""print the gpl notice to screen"""
	print """
Smileys Be Gone! version 1.0
Copyright (C) 2006 Neeraj Kumar and Robert Marianski
Smileys Be Gone! comes with ABSOLUTELY NO WARRANTY.
This is free software.
You are welcome to redistribute it under certain conditions.
See license for details.
	"""

# some utility functions
def getMag(vec):
	"""Returns the magnitude of the given vector.
	Examples:
	>>> getMag([1, 0])
	1.0
	>>> getMag([0, 2])
	2.0
	>>> getMag([3, 4])
	5.0
	"""
	assert(vec != None)
	ret = 0.0
	for i in vec: ret += i*i
	return sqrt(ret)

def getDir(from_, to_):
	"""Returns the direction from 'from_' to 'to_' as a vector.
	Examples:
	>>> getDir([1, 3], [5, 1])
	[4, -2]
	>>> getDir([0.1, 4.1, 3.8], [5.35, -2.4, 3.8])
	[5.25, -6.5, 0.0]
	"""
	return map(op.sub, to_, from_)

def magToVec(m, dir):
	"""Returns a vector in the same direction as 'dir' of length 'm'
	Examples:
	>>> magToVec(10, [1, 0])
	[10.0, 0.0]
	>>> magToVec(5, [3, 4])
	[3.0, 4.0]
	>>> magToVec(-1, [2, 1])
	[-0.89442719099991586, -0.44721359549995793]
	"""
	oldmag = getMag(dir)
	if oldmag == 0: oldmag = 1.0
	ret = []
	for i in dir:
		ret.append(i*m/oldmag)
	return ret

def vecToAngle(vec):
	"""Converts a vector to an angle between -180 and 180.
	Examples:
	>>> vecToAngle([0, 1])
	90.0
	>>> vecToAngle([-1, 0])
	180.0
	>>> vecToAngle([-1, 1])
	135.0
	>>> vecToAngle([1, 1])
	45.0
	>>> vecToAngle([1, 0])
	0.0
	"""
	assert len(vec) == 2
	return degrees(atan2(float(vec[1]), float(vec[0])))


def angleToVec(angle):
	"""Returns a unit-length vector corresponding to an angle of 'angle' degrees.
	Angles are measured counter-clockwise from 0 pointing right (in the +x direction).
	Examples:
	>>> angleToVec(0)
	(1.0, 0.0)
	>>> angleToVec(45)
	(0.70710678118654757, 0.70710678118654746)
	>>> angleToVec(-45)
	(0.70710678118654757, -0.70710678118654746)
	"""
	x, y = cos(radians(angle)), sin(radians(angle))
	return (x, y)

def tupmap(f, *tuples):
	"""apply a function f to the list of tuples and return a tuple
	>>> tupmap(op.add, (1,2,3), (1,2,3))
	(2, 4, 6)
	>>> tupmap(op.sub, (5,5), (1,1))
	(4, 4)
	>>> a = (3,4)
	>>> b = (2,2)
	>>> distance(a,b) == getMag(tupmap(op.sub, a, b))
	True
	>>> tupmap(avg, (1,2,3), (1,2,3), (1,2,3), (1,2,3))
	(1.0, 2.0, 3.0)
	>>> tupmap(op.neg, (-1,-2,-3))
	(1, 2, 3)
	"""
	return tuple(imap(f, *tuples))

def avg(*args):
	"""
	>>> avg(1,2)
	1.5
	>>> avg(-1, 1)
	0.0
	>>> avg(10, 20, 30, 40)
	25.0
	"""
	return sum(args) / float(len(args))

def avgTuples(tuples):
	""" 
	>>> avgTuples([(1,2)])
	[1.0, 2.0]
	>>> avgTuples([(1,2),(3,4),(5,6)])
	[3.0, 4.0]
	>>> avgTuples([(1,2),(2,3)])
	[1.5, 2.5]
	"""
	return map(avg, *tuples)

def distance(v1, v2):
	"""
	>>> distance([1, 2], [5, 2])
	4.0
	>>> distance([5, 2], [1, 2])
	4.0
	>>> distance([1, 2], [4, 6])
	5.0
	"""
	return getMag([i-j for i, j in zip(v1, v2)])

def combinedSeq(*args):
	"""Returns a single sequence containing all the elements of the passed in sequences
	>>> combinedSeq()
	()
	>>> combinedSeq(1)
	(1,)
	>>> combinedSeq([1])
	(1,)
	>>> combinedSeq([1], (2))
	(1, 2)
	>>> combinedSeq((1,3), (2,1,5))
	(1, 3, 2, 1, 5)
	>>> combinedSeq(1,2,3,[4,5,6], (7,8))
	(1, 2, 3, 4, 5, 6, 7, 8)
	"""
	ret = []
	for i in args:
		try:
			ret.extend(i)
		except Exception: ret.append(i)
	return tuple(ret)

def abstract():
	"""Place after a method body to declare that class abstract. Taken from 
	Peter Norvig's Python IAQ at: http://www.norvig.com/python-iaq.html"""
	import inspect
	caller = inspect.getouterframes(inspect.currentframe())[1][3]
	raise NotImplementedError(caller + ' must be implemented in subclass')

def make_rect(centerpos, size):
	""" return a Rect from centerpos and size tuples
	>>> make_rect((2,2), (4,4))
	<rect(0, 0, 4, 4)>
	>>> make_rect((1,1), (2,2))
	<rect(0, 0, 2, 2)>
	>>> make_rect((25,25), (2,2))
	<rect(24, 24, 2, 2)>
	>>> make_rect((55, 85), (25, 30))
	<rect(43, 70, 25, 30)>
	"""
	# easiest to make a rect and delegate the transform to that
	r = Rect(0,0,1,1)
	r.size = size
	r.center = centerpos
	return r

def cummap(mapop, redop, seq, init, *args):
	"""Cumulative map (or combined map/reduce).
	This takes a sequence and returns a sequence, where each element is constructed from applying
	reduce to the sequence so far. The value from applying mapop to each list element (along
	with any other *args) is then reduced using redop, and this value is appened to the output list.
	Examples:
	>>> cummap(lambda x: x, op.add, [1, 2, 3], 0)
	[1, 3, 6]
	>>> cummap(lambda x: 2*x, op.mul, [3, 2, 1], 1)
	[6, 24, 48]
	"""
	ret = []
	cur = init
	for i in range(len(seq)):
		val = mapop(seq[i], *args)
		cur = redop(cur, val)
		ret.append(cur)
	return ret

def choiceWithProbs(seq, probs):
	"""Returns an element from seq chosen using the probability distribution in probs.
	probs and seq should have the same length. Works if the given probs are negative as well
	as if they're all equal to 0"""
	# normalize the probabilities first and compute the cdf
	assert len(seq) == len(probs)
	m = min(probs)
	probs = [p+m for p in probs]
	mag = sum(probs)
	if mag == 0: return choice(seq)
	probs = cummap(lambda x: x, op.add, [p/mag for p in probs], 0.0)
	# generate a random number and use that to determine which bin to choose
	r = random()
	for p, el in zip(probs, seq):
		if r <= p: return el
	assert "We should have returned already"

if __name__ == "__main__":
	import doctest
	doctest.testmod()
