# Smileys Be Gone!
# A game with the goal to make smileys disappear
# Copyright (C) 2006 Neeraj Kumar and Robert Marianski

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import pygame
from pygame.locals import *
from math import *
from random import *
from gamestructs import *
from gameutils import *
from gamegraphics import *
from gameaudio import *
import operator as op


def displayImageAndWait(fname, screen, waitamt=500):
	"""Displays the a given image and waits until the user presses or clicks something"""
	import pygame.event

	screen.blit(pygame.image.load(fname), (0, 0))
	pygame.display.flip()

	while 1:
		# if the user clicks or presses anything, return
		if pygame.event.get(QUIT): raise SystemExit
		if pygame.event.get([MOUSEBUTTONDOWN, KEYDOWN]): return

		# kill all other events
		junk = pygame.event.get()

		# wait
		pygame.time.wait(waitamt)
			
class Game(object):
	"""The main game class that manages everything"""

	# enum for the different placeable objects
	NUM_TOOLS = 3
	SPRING, MONEY, DOGSHIT = range(NUM_TOOLS)
	ToolTypes = {SPRING: ("springbutton_*.png", (640, 100)),
	              MONEY: ("moneybutton_*.png", (700, 100)),
				  DOGSHIT: ("dogshitbutton_*.png", (760, 100))}

	# enum and list for death types
	NUM_DEATH_TYPES = 6
	HITNRUN, BUILDINGSPLAT, TRAINSPLAT, SENT, CARCRASH, TRAINCRUSH = range(NUM_DEATH_TYPES)
	DeathTypes = {HITNRUN: ("Hit And Runs: %d", (605, 240)),
	              BUILDINGSPLAT: ("Building Splatters: %d", (605, 260)),
				  TRAINSPLAT: ("Train Splatters: %d", (605, 280)),
				  SENT: ("Sent Into The Heavens: %d", (605, 300)),
				  CARCRASH: ("Car Crashes: %d", (605, 320)),
				  TRAINCRUSH: ("Cars Crushed By Trains: %d", (605, 340))}

	# enum for mission success or failure
	IN_PROGRESS, SUCCESS, FAILURE = range(3)

	# enum for game difficulty
	EASY, MEDIUM, HARD = range(3)

	# difficulty strings for displaying on screen
	difficultyStrings = {EASY:   "Easy",
	                     MEDIUM: "Medium",
	                     HARD:   "Hard"}

	# String and position of goals to list
	goalStringPos = (605, 450)
	goalStrings = {HITNRUN: "Smileys Run Over: %d/%d",
	               BUILDINGSPLAT: "Smiley Building Fodder: %d/%d",
								 TRAINSPLAT: "Smileys Derailed: %d/%d",
								 SENT: "Smileys Sent: %d/%d",
								 CARCRASH: "Car Crashes: %d/%d",
								 TRAINCRUSH: "Cars Derailed: %d/%d"}

	# amount of scaling of number of deaths based on level goal
	deathScales = {HITNRUN: 2.0,
	               BUILDINGSPLAT: 3.0,
								 TRAINSPLAT: 1.0,
								 SENT: 1.0,
								 CARCRASH: 1.0,
								 TRAINCRUSH: 1.5}

	# the path to level files
	levelfnames = 'levels/Level%04d.dat'

	# the framerate we want our game to run at
	framerate = 30

	# the spawn probability for cars (should be applied as follows): 
	#    if random() < carSpawnProb*(1.0 - curCars/maxCars): spawnCar()
	carSpawnProb = 0.1

	# the spawn probability for peds (used in the same way as the carSpawnProb)
	pedSpawnProb = 0.01

	# the minimum speed difference between cars for it to be marked dead (pixels/ms)
	carCrashMinDif = 0.01

	# the amount a building pushes peds out (pixels)
	buildingPush = 1.0

	# the minspeed for a flying ped to be killed if hit on a building
	buildingMinSpeed = 0.1

	# the time limit for each level (this remains the same)
	timelimit = 180000

	def __init__(self, screen, winsize, gamesize):
		"""Creates a new game"""
		# constants for the duration of the game
		self.screen = screen
		self.screensize = tuple(winsize)
		self.screenRect = Rect(0, 0, *self.screensize)
		self.gamesize = tuple(gamesize)
		self.gameRect = Rect(0, 0, *self.gamesize)
		self.rightpanelsize = (winsize[0]-gamesize[0], gamesize[1])
		self.rightpanelbg = Sprite()
		self.rightpanelbg.image = pygame.Surface(self.rightpanelsize)
		self.rightpanelbg.image.fill((255, 205, 150))
		self.rightpanelbg.rect = Rect(self.gamesize[1], 0, *self.rightpanelsize)

		# test if we have sound
		if pygame.mixer.get_init():
			self.sound = 1
			self.music = 1
		else:
			self.sound = 0
			self.music = 0

		self.levelmusic = None

		# clear out the level
		self.resetLevel()

	def resetLevel(self):
		import pygame.sprite as s
		"""Clears out everything in the level"""
		# level-related data
		self.background = pygame.Surface(self.screensize)
		self.background.fill((0,0,0))
		self.curlevel = -1
		if self.levelmusic:
			self.levelmusic.stop()

		# state management
		self.goals = []
		self.availableObjects = [0] * Game.NUM_TOOLS
		self.lastDragged = None
		self.pressed = None
		self.lastLightQuanta = None
		self.lastTime = 0
		self.curTime = None
		self.curTool = -1
		self.timeleft = 0
		self.mouseObject = SpriteRect(Rect(0,0,1,1))

		# lists of non-graphics objects
		self.nodes = None
		self.edges = None
		self.carSpawns = []
		self.deaths = {}
		for type in Game.DeathTypes.keys(): self.deaths[type] = {}
		
		# lists of graphics objects
		# game objects
		self.lights = s.RenderUpdates()
		self.cars = s.RenderUpdates()
		self.trains = s.RenderUpdates()
		self.buildings = s.RenderUpdates()
		self.peds = s.RenderUpdates()
		self.walkingPeds = s.RenderUpdates()
		self.flyingPeds = s.RenderUpdates()
		self.deadPeds = s.RenderUpdates()
		self.springs = s.RenderUpdates()
		self.forceObjects = s.RenderUpdates()
		self.sidewalks = s.RenderUpdates()
		self.trainSpawns = s.RenderUpdates()
		self.flames = s.RenderUpdates()

		# game interface objects
		self.rightpanel = s.RenderUpdates()
		self.buttons = s.RenderUpdates()

		# the drawing order
		self.todraw = [self.buildings, self.springs, self.forceObjects, self.deadPeds, 
		               self.walkingPeds, self.cars, self.trains, self.lights, self.flyingPeds,
					   self.trainSpawns, self.flames, self.rightpanel, self.buttons]
		self.rightpanel.add(self.rightpanelbg)

		# the sound controls, if we have sound
		self.soundcontrol_group = s.RenderUpdates()
		self.todraw.append(self.soundcontrol_group)

		# add the appropriate sound sprites if we have sound enabled
		if self.music:
			def music_notifier(self=self):
				self.levelmusic.updateVolume()
			self.music_control = VolumeControl(BaseSound.MUSIC, (670, 570), notifyfn=music_notifier)
			self.soundcontrol_group.add(self.music_control)
		
		if self.sound:
			self.sound_control = VolumeControl(BaseSound.SOUND, (730, 570))
			self.soundcontrol_group.add(self.sound_control)


	def getLevelParams(self, difficulty, nodes, edges):
		"""Returns the various level parameters given a level number and difficulty"""
		# figure out the total length of roads and sidewalks
		lens = {}
		for type in [PED, ROAD]:
			curlen = 0
			for el in combinedSeq(nodes[type], edges[type]):
				curlen += el.size
			lens[type] = curlen

		# max number of cars is some constant factor of this, +/- a random 10 percent
		maxcars = int(lens[ROAD] / 200)
		maxcars += randint(int(-maxcars/10), int(maxcars/10))
		if maxcars < 10: maxcars = 10
		if maxcars > 100: maxcars = 100

		# init number of cars is just some percentage of this, based on difficulty level
		initcars = int(maxcars/(difficulty*difficulty + 2))

		# max number of peds is dependent on both sidewalks length and on difficulty
		maxpeds = int(lens[PED] / (100 * (difficulty+1)))
		maxpeds += randint(int(-maxpeds/10), int(maxpeds/10))
		if maxpeds < 15: maxpeds = 15
		if maxpeds > 200: maxpeds = 200

		# init number of peds is just like init number of cars
		initpeds = int(maxpeds/(difficulty*difficulty + 2))

		# number of tools is dependent on difficulty
		numtools = [5, 3, 3]
		# apply difficulty - drop 1/3 for each level
		numtools = [i-(i*difficulty/3) for i in numtools]

		# goals are dependent on difficulty
		possible_goals = Game.DeathTypes.keys()[:]
		# randomize selection
		# all goals have an equal probability of being chosen
		shuffle(possible_goals)
		# remove train goals if level has no trains
		if not len(edges[TRAIN]):
			for g in (Game.TRAINSPLAT, Game.TRAINCRUSH):
				possible_goals.remove(g)
		# remove building goal if no buildings
		if not len(nodes[BUILDING]):
			possible_goals.remove(Game.BUILDINGSPLAT)
		# TODO also need to check to see if train intersects road at all
		goaltype = possible_goals.pop()
		ndeaths = 20*(difficulty+1)
		# scale pedestrian type goals
		assert goaltype in (Game.HITNRUN, Game.BUILDINGSPLAT, Game.TRAINSPLAT, Game.SENT, Game.CARCRASH, Game.TRAINCRUSH)
		# scale deaths
		if goaltype in (Game.HITNRUN, Game.BUILDINGSPLAT, Game.TRAINSPLAT, Game.SENT):
			ndeaths += maxpeds*Game.deathScales[goaltype]/2
		else:
			ndeaths += maxcars*Game.deathScales[goaltype]/2
		goals = [DeathGoal(self.deaths[goaltype], ndeaths, goalstr=Game.goalStrings[goaltype])]
		# on a hard difficulty, we add 2 goals
		if difficulty == Game.HARD:
				goaltype = possible_goals.pop()
				goals.append(DeathGoal(self.deaths[goaltype], ndeaths, goalstr=Game.goalStrings[goaltype]))
		return (numtools, initpeds, maxpeds, initcars, maxcars, goals)
		
	def loadLevel(self, levelnum, difficulty):
		"""Loads the given level number"""
		from bz2 import BZ2File
		from pickle import load

		self.screen.fill((0, 0, 0))
		pygame.display.flip()

		# load the given file and then make sure we got valid data
		levelname = self.fnames[levelnum % len(self.fnames)]
		(nodes, edges) = load(open(levelname, 'rb'))
		assert len(nodes) == 4
		assert len(edges) == 4
		for type in [TRAIN, ROAD, PED, BUILDING]:
			assert type in nodes
			assert type in edges
		# reset the level
		self.resetLevel()
		self.curlevel = levelnum
		self.difficulty = difficulty
		# now set our nodes and edges to those loaded
		self.nodes = nodes
		self.edges = edges
		# set our background
		try:
			f = BZ2File(levelname.replace('.dat', '.bgi'), 'r')
			sbuffer = load(f)
			bg = pygame.image.fromstring(sbuffer, self.gamesize, "RGB")
			self.background = bg
		except Exception, e:
			print "Exception: ", e
		self.screen.blit(self.background, (0, 0))

		# find the spawn points
		for n in nodes[TRAIN]:
			# all train nodes are spawn points
			self.trainSpawns.add(TrainSpawn(n, self.gamesize)) 
		for n in nodes[ROAD]:
			if n.isSpawn: self.carSpawns.append(CarSpawn(n))

		# add all the sidewalk nodes and edges to the list of sidewalks, for mouse-click purposes
		for o in combinedSeq(nodes[PED], edges[PED]):
			sr = SpriteRect(o.getRect())
			# store a pointer to the actual node or edge
			sr.object = o
			self.sidewalks.add(sr)
			
			# also add a 'score' field to each item
			o.score = 0.0

		# generate various parameters based on level number and difficulty
		params = self.getLevelParams(difficulty, nodes, edges)
		self.availableObjects, initpeds, self.maxpeds, initcars, self.maxcars, self.goals = params

		# go through all the edges and spawn people on them as needed
		if len(edges[PED]) > 0:
			# also put in ped lights for ones that are crosswalks
			for e in edges[PED]:
				e.lights = []
				# bloat the sidewalk edges to get them to go to the center of the nodes
				e.size += e.nodes[0].size
				# TODO ped lights are annoying, so decide whether to keep them or not
				continue
				try:
					if e.crosswalk:
						if e.orientation == HORIZONTAL: offsets = (e.size/2, 0)
						else: offsets = (0, e.size/2)
						# place the lights
						for pos in [map(op.sub, e.pos, offsets), map(op.add, e.pos, offsets)]:
							pl = PedLight(1-e.orientation, pos)
							self.lights.add(pl)
							# also save pointers to the light in the edge itself, so that  can access it
							e.lights.append(pl)
				except Exception: continue
			self.spawnPeds(1)

		# add all the building nodes and edges to the list of buildings, for collision purposes
		for o in combinedSeq(nodes[BUILDING], edges[BUILDING]):
			self.buildings.add(SpriteRect(o.getRect()))

		# add all the traffic lights
		for e in edges[ROAD]:
			# get the offsets at which to place the lights
			if e.orientation == HORIZONTAL: offsets = (e.size/2, 0)
			else: offsets = (0, e.size/2)
			e.lights = []
			# place the lights
			for pos in [map(op.sub, e.pos, offsets), map(op.add, e.pos, offsets)]:
				# remember that vertical roads have horizontal lights, and vice-versa
				cl = CarLight(1-e.orientation, pos)
				self.lights.add(cl)
				# also save pointers to the light in the edge itself, so that cars can access it
				e.lights.append(cl)

		# spawn cars in the roads (so we don't have to wait for them to slowly come in)
		while initcars > 0 and len(edges[ROAD]) > 0:
			# pick a random road and try to spawn there
			e = choice(edges[ROAD])
			if self.spawnCarOnEdge(e): initcars -= 1

		# create the buttons
		for k, v in Game.ToolTypes.iteritems():
			fnameGlob, pos = v
			def callback(self=self, tool=k):
				self.setCurTool(tool)
			self.buttons.add(Button(fnameGlob, pos, k, callback))
		self.setCurTool(Game.SPRING)

		# start the clock
		self.starttime = pygame.time.get_ticks()

		# start playing music
		if self.music:
			self.levelmusic = LevelMusic(LevelMusic.GAME)
			self.levelmusic.play(repeat=1)

		# at the very end, we flip our display to show the background
		pygame.display.flip()


	def addDeath(self, victim, killer):
		"""Adds a death where 'killer' killed 'victim'"""
		# figure out what type of death it was
		type = -1
		# also keep track of how many times that person was bounced
		numBounces = 0
		if victim in self.peds:
			# victim was a ped
			if victim.state == Ped.DEAD: return
			if killer in self.cars: type = Game.HITNRUN
			elif killer in self.trains: type = Game.TRAINSPLAT
			elif killer in self.buildings: type = Game.BUILDINGSPLAT
			elif killer == self: type = Game.SENT
			numBounces = victim.numBounces
		elif victim in self.cars:
			# victim was a car
			if victim.state == Car.DEAD: return
			if killer in self.trains: type = Game.TRAINCRUSH
			elif killer in self.cars: type = Game.CARCRASH
			# start the flames animation
			def callback(car=victim):
				car.kill()
			self.flames.add(Flames(victim.pos, self.curTime, callback=callback))
			# play the carcrash sound
			if self.sound:
				SoundEffect(SoundEffect.CARCRASH, victim.pos[0])

		# increment that death counter
		if type >= 0: 
			if numBounces in self.deaths[type]:
				self.deaths[type][numBounces] += 1
			else:
				self.deaths[type][numBounces] = 1

		# finally, kill the victim
		victim.deathTime = self.curTime
		victim.kill()

	def spawnCarOnEdge(self, e):
		"""Tries to spawn a car on the given edge. If it succeeds, returns 1, else returns 0"""
		import pygame.sprite as s
		# create a new car at a random position on this edge
		c = Car(e, (random()*0.8) + 0.1, choice([FORWARD, BACKWARD]))
		# if it collides with any existing cars, return 0
		if s.spritecollideany(c, self.cars): 
			c.kill()
			return 0
		# if it doesn't, then add it to the list of cars and return 1
		self.cars.add(c)
		return 1

	def spawnPeds(self, force=0):
		"""Spawns peds on sidewalks. If force is true, then definitely spawns a ped"""
		if len(self.peds) >= self.maxpeds: return
		for e in self.edges[PED]:
			if force or random() < Game.pedSpawnProb*(1.0 - len(self.peds)/float(self.maxpeds)):
				p = Ped(e, random(), choice([FORWARD, BACKWARD]), (self.walkingPeds, self.flyingPeds, self.deadPeds), self.nodes[PED], self.edges[PED])
				self.peds.add(p)
				self.walkingPeds.add(p)
			
	def setCurTool(self, tool):
		for b in self.buttons: b.setState(b.NORMAL)
		# if we have more of this tool to place, then place it
		if self.availableObjects[tool] > 0:
			for b in self.buttons:
				if b.obj == tool:
					b.setState(b.ACTIVE)
					break
			self.curTool = tool

	def click(self, clickevents, upevents):
		"""Processes mouse commands"""
		for e in clickevents:
			# set the mouseobject's rect to where the user just clicked
			self.mouseObject.rect = Rect(e.pos[0], e.pos[1], 1, 1)
			def coll(grp, dokill=0):
				return pygame.sprite.spritecollide(self.mouseObject, grp, dokill)
			#print "Clicked mouse button %d at: " % e.button, e.pos
			if e.button == 3:
				#If they right-clicked, then just search the list of placeable objects, killing the clicked obj
				coll(self.forceObjects, 1)
				coll(self.springs, 1)
				# if they right-clicked a light, set it back to normal timings
				ret = coll(self.lights)
				for l in ret:
					l.resetBackToTimings(self.curTime)
					return
			elif e.button == 1:
				self.pressed = e.pos[:]
				# if they left-clicked, then figure out what they clicked on
				# first check if they clicked on a sound control
				if self.sound:
					for soundcontrol in coll(self.soundcontrol_group):
						soundcontrol.click()
						return
				ret = coll(self.buttons)
				for button in ret:
					# they clicked a button
					button.click()
					return
				ret = coll(self.lights)
				for l in ret:
					# they clicked on a light
					l.click(self.curTime)
					return
				ret = coll(self.trainSpawns)
				for ts in ret:
					# they clicked on a train spawn
					t = ts.spawn(self.curTime, clicked=1)
					if t: 
						self.trains.add(t)
						if self.sound:
							SoundEffect(SoundEffect.TRAIN, t.pos[0])
					return
				ret = coll(self.sidewalks)
				for sidewalk in ret:
					# they clicked somewhere on a sidewalk
					if self.availableObjects[self.curTool] < 1: continue
					def changeAmount(incr, l=self.availableObjects, t=self.curTool):
						self.availableObjects[t] += incr
					if self.curTool == Game.MONEY:
						self.forceObjects.add(ForceObject(ForceObject.MONEY, e.pos, sidewalk.object, changeAmount))
					elif self.curTool == Game.DOGSHIT:
						self.forceObjects.add(ForceObject(ForceObject.DOGSHIT, e.pos, sidewalk.object, changeAmount))
					self.setCurTool(self.curTool)
					return

		if self.pressed and len(upevents) > 0:
			e = upevents[0]
			# set the mouseobject's rect to where the user just let go
			self.mouseObject.rect = Rect(self.pressed[0], self.pressed[1], 1, 1)
			# if we're trying to placing a spring, then figure out if it's legal
			if self.curTool == Game.SPRING:
				ret = pygame.sprite.spritecollide(self.mouseObject, self.sidewalks, 0)
				for sidewalk in ret:
					# they clicked on this sidewalk, so figure out the offset
					offset = map(op.sub, e.pos, self.pressed)
					def changeAmount(incr, l=self.availableObjects, t=self.curTool):
						self.availableObjects[t] += incr
					self.springs.add(Spring(self.pressed, offset, sidewalk.object, changeAmount))
			self.pressed = None

	def keypress(self, keyevents):
		"""Processes keyboard commands"""
		for ke in keyevents:
			k = ke.key
			if k in (K_ESCAPE, K_q): return 1
			elif k == K_s: self.setCurTool(Game.SPRING)
			elif k == K_m: self.setCurTool(Game.MONEY)
			elif k == K_d: self.setCurTool(Game.DOGSHIT)
		return 0

	def changeLights(self, curTime):
		"""Changes the lights"""
		# first figure out the current light quanta
		quanta = (curTime/Light.lightQuanta) % sum(CarLight.times)
		# if the light quanta is still the same, we don't need to do anything
		if quanta == self.lastLightQuanta: return
		
		# otherwise, the light quanta has changed
		for l in self.lights:
			# if we're not allowed to change this light, continue
			if curTime < l.mintime: continue
			l.setQuanta(quanta, curTime)
		
		# finally, remember what quanta we're at
		self.lastLightQuanta = quanta

	def spawnCars(self):
		"""Spawns new cars as needed"""
		if len(self.carSpawns) == 0: return
		# first see how many cars we have
		# if we have enough, just leave
		if len(self.cars) >= self.maxcars: return
		# otherwise, try to spawn one at each spawn point
		for cs in self.carSpawns:
			if random() < Game.carSpawnProb*(1.0 - len(self.cars)/float(self.maxcars)):
				cs.spawn(self.cars)

	def applyEvents(self):
		"""Gets and applies user-generated events"""
		import pygame.event as e
		# first see if the user wanted to quit
		# a quit event means we should exit right away
		if e.get(QUIT): raise SystemExit

		# get all mouse click-events on primary or secondary button
		clickevents = [i for i in e.get([MOUSEBUTTONDOWN]) if i.button in (1, 3)]

		# get all mouse drag-events
		dragevents = [i for i in e.get([MOUSEMOTION]) if i.buttons[0] == 1]
		# save the location where the mouse was last dragged to
		if dragevents: self.lastDragged = dragevents[-1].pos

		# get mouse-release events only for the first mouse button
		upevents = [i for i in e.get([MOUSEBUTTONUP]) if i.button == 1]
		self.click(clickevents, upevents)

		# get all keyboard-events
		keyevents = e.get([KEYDOWN])
		ret = self.keypress(keyevents)
		if ret:
			self.resetLevel()
			return 1

		# drain the event queue of everything else
		junk = e.get()
		return 0
	
	def doRegularEvents(self):
		"""Does regular events"""
		# change all the lights
		self.changeLights(self.curTime)

		# spawn cars
		self.spawnCars()

		# spawn pedestrians
		self.spawnPeds()

		# spawn trains
		for ts in self.trainSpawns:
			t = ts.spawn(self.curTime)
			if t: 
				self.trains.add(t)
				if self.sound:
					SoundEffect(SoundEffect.TRAIN, t.pos[0])

	def moveObjects(self):
		"""Moves objects around"""
		# the length of the current timestep
		timestep = self.curTime-self.lastTime
		if timestep < 1: return

		# because cars react to each other, we can't simply move them like other objects
		# we first need to adjust their accelerations based on current positions and velocities
		for c in self.cars: c.adjustAcceleration(timestep)

		# now we can move everything around just by calling move() on each one--ah, inheritance rules!
		for obj in combinedSeq(self.cars, self.flyingPeds, self.walkingPeds, self.trains):
			obj.move(timestep)

	def doCollisions(self):
		"""Does collisions between objects"""
		import pygame.sprite as s
		# car-car
		crashes = s.groupcollide(self.cars, self.cars, 0, 0)
		for c1, c2 in crashes.iteritems():
			if len(c2) == 0: continue
			c2 = c2[0]
			# first check to see if c1 is already dead
			if c1.state == Car.DEAD: continue
			# if the speed difference is not much, then ignore the crash
			dif = [v2-v1 for v1, v2 in zip(c1.vel, c2.vel)]
			speedDif = getMag(dif)
			if speedDif < Game.carCrashMinDif: continue
			# otherwise, mark both cars as dead
			self.addDeath(c1, c2)
			self.addDeath(c2, c1)

		# car-train
		crashes = s.groupcollide(self.cars, self.trains, 0, 0)
		for c, t in crashes.iteritems():
			if len(t) == 0: continue
			if c.state == Car.DEAD: continue
			t = t[0]
			# car got 0wn3d by a train
			self.addDeath(c, t)

		# person-car/train
		for obstacle in [self.cars, self.trains]:
			collisions = s.groupcollide(self.peds, obstacle, 0, 0)
			for p, o in collisions.iteritems():
				if len(o) == 0: continue
				# person got 0wn3d by a car/train
				o = o[0]
				self.addDeath(p, o)

		# walking ped-building
		hits = s.groupcollide(self.walkingPeds, self.buildings, 0, 0)
		for p, b in hits.iteritems():
			if len(b) == 0: continue
			b = b[0]
			# just push the person out a little bit
			dir = map(op.sub, p.pos, b.rect.center)
			push = magToVec(Game.buildingPush, dir)
			p.pos = map(op.add, p.pos, push)

		# flying ped-building
		splats = s.groupcollide(self.flyingPeds, self.buildings, 0, 0)
		for p, b in splats.iteritems():
			if len(b) == 0: continue
			b = b[0]
			# if the person's not going that fast, then bounce them off the building
			if getMag(p.vel) < Game.buildingMinSpeed:
				# figure out if they collide on a vertical face or horizontal face of the building
				if p.pos[0] >= b.rect.left and p.pos[0] <= b.rect.right:
					# they bounced off vertical face, so reverse their y-vel
					p.vel[1] = -p.vel[1]
					# also push them out of the building
					if p.pos[1] < b.rect.centery:
						p.pos[1] -= Game.buildingPush
					else:
						p.pos[1] += Game.buildingPush
				else:
					# they bounced off the horizontal face, so reverse their x-vel
					p.vel[0] = -p.vel[0]
					# also push them out of the building
					if p.pos[0] < b.rect.centerx:
						p.pos[0] -= Game.buildingPush
					else:
						p.pos[0] += Game.buildingPush
				p.numBounces += 1
			else:
				# they were flying too fast, so they got splatted on a building
				self.addDeath(p, b)

		# flying ped-ped
		tagteams = s.groupcollide(self.flyingPeds, self.peds, 0, 0)
		for p1, p2 in tagteams.iteritems():
			if len(p2) == 0: continue
			# person got hit by a flying ped. average their velocities and make them all fly
			v = [i for i in p1.vel]
			count = 1
			for p in p2:
				v = [i1+i2 for i1, i2 in zip(v, p.vel)]
				count += 1.0
			v = [i/count for i in v]
			p1.vel = v[:]
			for p in p2:
				p.state = Ped.FLYING
				p.vel = v[:]

		# ped-spring
		bounced = s.groupcollide(self.peds, self.springs, 0, 0)
		for p, sp in bounced.iteritems():
			if len(sp) == 0: continue
			sp = sp[0]
			if sp.state == Spring.BUSY: continue
			# person landed on a spring: time to fly!
			applied = sp.applyToPerson(p, self.curTime)
			if applied and self.sound:
				SoundEffect(SoundEffect.SPRING, sp.pos[0])

		# ped-force object: simply kill that object
		picked = s.groupcollide(self.peds, self.forceObjects, 0, 1)
		for p, fo in picked.iteritems():
			if not fo: continue
			if p.state == p.DEAD: continue
			fo = fo[0]
			if fo.type == ForceObject.MONEY:
				if self.sound:
					SoundEffect(SoundEffect.MONEY, fo.pos[0])

		# flying ped-gamesize
		for p in self.flyingPeds:
			# if the person doesn't collide with the gamerect, he must be outside...
			# ...and there's no way back in, so he's been SENT (INTO THE HEAVENS!)
			if not p.rect.colliderect(self.gameRect): self.addDeath(p, self)

	def clearAndDraw(self):
		"""Clears relevant areas and draws to the screen"""
		# clears old stuff
		for objs in self.todraw:
			objs.clear(self.screen, self.background)

		# update our current tool to see if we have any left
		self.setCurTool(self.curTool)

		# keep a list of all the rects we're gonna wanna redraw
		rectlist = []
		for objs in self.todraw:
			# call update on the objects, in case they need to do some internal bookkeeping
			objs.update(self.curTime)
			# draw that object, saving the list of changed areas
			rectlist.extend(objs.draw(screen))

		# convenience func to write a string to the screen
		def drawString(s, color, pos): self.screen.blit(font.render(s, True, color), pos)

		# write the game name
		font = pygame.font.Font("gamefont.ttf", 20)
		drawString("Smileys Be Gone!", (0, 0, 0), (620, 20))
		font = pygame.font.Font("gamefont.ttf", 14)

		# write out the current level and difficulty
		drawString("Level: %d" % (self.curlevel+1), (0, 0, 255), (605, 160))
		difficultyString = Game.difficultyStrings[self.difficulty % 3]
		drawString("Difficulty: %s" % difficultyString, (0, 0, 255), (605, 180))

		# write out the current counts on our various tools
		for k, v in Game.ToolTypes.iteritems():
			fname, pos = v
			pos = [pos[0]-3, pos[1]+25]
			s = "%d" % self.availableObjects[k]
			if self.curTool == k:
				color = (0, 0, 208)
			else:
				color = (100, 150, 240)
			drawString(s, color, pos)

		# write out the deathcounts
		maxbounces = 0
		for type, v in Game.DeathTypes.iteritems():
			s, pos = v
			# count the total number of deaths of this type (i.e. from any number of bounces)
			total = 0
			for nbounces, ndeaths in self.deaths[type].iteritems():
				total += ndeaths
				if nbounces > maxbounces: maxbounces = nbounces
			drawString(s % total, (255, 0, 0), pos)

		# write out our current goals
		pos = Game.goalStringPos
		for g in self.goals:
			drawString(str(g), (255, 0, 255), pos)
			pos = map(op.add, pos, (0, 20))

		# write out the current time left
		font = pygame.font.Font("gamefont.ttf", 20)
		mins = int(self.timeleft/60000)
		secs = int((self.timeleft-(mins*60000))/1000)
		if mins == 0 and secs <= 20:
			color = (255, 0, 0)
		else:
			color = (0, 0, 0)
		drawString("%d:%02d left" % (mins, secs), color, (657, 500))

		# finally, update our screen 
		pygame.display.update(rectlist)

	def updateGoals(self):
		"""Updates our goals and returns True if we achieved them"""
		results = []
		for g in self.goals:
			results.append(g.passedGoal())
		all = reduce(lambda x, y: x and y, results, True)
		if all:
			return Game.SUCCESS
		self.timeleft = self.starttime + Game.timelimit - self.curTime
		if self.timeleft < 0:
			return Game.FAILURE
		return Game.IN_PROGRESS

	def main(self, levelnum, difficulty):
		"""Main loop of the game"""
		#print "Starting game at level %s and difficulty %d" % (levelnum, difficulty)
		self.loadLevel(levelnum, difficulty)
		quit = 0
		self.lastTime = pygame.time.get_ticks()
		while not quit:
			self.curTime = pygame.time.get_ticks()
			# get and apply (user-generated) events
			quit = self.applyEvents()
			if quit: return
			# apply regular events
			self.doRegularEvents()
			# move objects
			self.moveObjects()
			# do collisions
			self.doCollisions()
			# now draw everything
			self.clearAndDraw()
			# update our goals 
			status = self.updateGoals()
			if quit or status == Game.FAILURE:
				displayImageAndWait('images/failure.png', self.screen)
				self.loadLevel(self.curlevel, difficulty)
				self.curTime = pygame.time.get_ticks()
			elif status == Game.SUCCESS:
				displayImageAndWait('images/success.png', self.screen)
				newlevel = (self.curlevel + 1) % len(self.fnames)
				self.loadLevel(newlevel, difficulty)
				self.curTime = pygame.time.get_ticks()
			# save the current time and sleep for a bit (because we have a busy timestep ahead =))
			self.lastTime = self.curTime
			newtime = pygame.time.get_ticks()
			elapsed = newtime-self.curTime
			if (elapsed < Game.framerate):
				pygame.time.wait(Game.framerate-elapsed)

	def createLevelStrings(self):
		"""create the different level names by listing the levels directory"""
		import os
		# get the level number from the file name
		# Level0001.dat -> int(1)
		parseLevelNum = lambda s:int(s[5:9])
		fnames = [f for f in os.listdir('levels') if f.endswith('.dat')]
		fnames.sort(key=parseLevelNum)
		# sort the list of fnames here
		# then when generating the second list, the order will match
		fnamestrings = ['Level %d' % parseLevelNum(f) for f in fnames]
		fnames = [os.path.join('levels', f) for f in fnames]
		return (fnames, fnamestrings)

	def mainLoop(self):
		"""The main loop which controls the menu system, etc."""
		import pygame.event as e
		import pygame.sprite as s
		# the different game interface states
		MAINMENU, HELP, PLAYGAME = range(3)
	
		# create the background
		background = pygame.Surface(self.screensize)
		background.fill((0,0,0))

		# start the background music
		menumusic = None
		if self.music:
			menumusic = LevelMusic(LevelMusic.MENU)
			menumusic.play(repeat=1)

		options = s.RenderUpdates()
		titles = s.RenderUpdates()
		# create the game name on the top
		# these are not text buttons, but we can reuse them for drawing to the screen
		gamename = TextButton("Smileys Be Gone!", font=pygame.font.Font("gamefont.ttf", 60), centerx=400, centery=130)
		production = TextButton("A Cornflower Blue Production", font=pygame.font.Font("gamefont.ttf", 30), color=(100,150,240), centerx=400, centery=200)
		titles.add(text for text in (gamename, production))
		for opt, retstate, (x,y) in zip(["Start Game", "Help", "Quit"],
			                              [PLAYGAME, HELP, QUIT],
														        [(400,300), (400,450), (400,500)]):
			tb = TextButton(opt, onclick=lambda s=retstate: s, centerx=x, centery=y)
			options.add(tb)

		# create the incrementing buttons
		self.fnames, fnamestrings = self.createLevelStrings()
		levelButton = IncrementingText(fnamestrings, inccolor=(100,150,240), centerx=400, centery=350)
		difficultButton = IncrementingText('Easy Medium Hard'.split(), inccolor=(100,150,240), centerx=400, centery=400)
		options.add(btn.getClickableSprites() for btn in (difficultButton, levelButton))

		# the animated random smileys
		smileys = s.RenderUpdates()
		for i in range(10):
			smileys.add(RandomSmiley(self.screensize, [smileys, options, titles]))
		# quick hack to make them start appearing faster
		for i in range(40):
			smileys.update()

		# the sprite-rect we'll use checking mouse clicks
		mouse = SpriteRect(Rect(0,0,1,1))

		while 1:
			state = MAINMENU
			# get the user's events and set the state as to what the user wants to do
			# window close
			quitevents = e.get(QUIT)
			if quitevents: return
			# pressed a key
			# 'q',escape -> quit
			# 's', 'n' -> start game
			# 'h', 'F1' -> help screen
			keyevents = e.get(KEYDOWN)
			for keyevent in keyevents:
				key = keyevent.key
				if key in [K_q, K_ESCAPE]: return
				elif key in [K_s, K_n]: state = PLAYGAME
				elif key in [K_h, K_F1]: state = HELP

			# see if they clicked a button
			# delegate to the sprite classes, which return the next state
			clickevents = e.get(MOUSEBUTTONDOWN)
			for click in clickevents:
				mouse.rect.center = click.pos
				ret = s.spritecollide(mouse, options, 0)
				for button in ret:
					state = button.click()
					if state == QUIT: return

			# see if they mouse-overed a button
			mouseovers = e.get([MOUSEMOTION])
			if mouseovers:
				for button in options: button.setState(Button.NORMAL)
			for mo in mouseovers:
				mouse.rect.center = mo.pos
				ret = s.spritecollide(mouse, options, 0)
				for button in ret:
					button.setState(Button.ACTIVE)

			# drain the event queue of everything else
			junk = e.get()

			# use the state to decide what to do
			if state == HELP:
				displayImageAndWait('images/help.png', self.screen)
			elif state == PLAYGAME:
				if menumusic:
					menumusic.stop(400)
				difficulty = difficultButton.i
				self.main(levelButton.i, difficulty)
				if self.music:
					menumusic = LevelMusic(LevelMusic.MENU)
					menumusic.play()

			# update the smileys
			smileys.update()

			self.screen.blit(background, (0,0))
			todraw = [smileys, options, titles]
			for obj in todraw:
				obj.clear(self.screen, background)
			rectlist = sum((objs.draw(screen) for objs in todraw), [])
			pygame.display.update()

			# wait a bit
			pygame.time.wait(30)


if __name__ == "__main__":
	from gameutils import print_license
	print_license()
	pygame.init()
	pygame.font.init()
	winsize = (800, 600)
	gamesize = (600, 600)
	screen = pygame.display.set_mode(winsize)
	pygame.display.set_caption("Smileys Be Gone!")
	game = Game(screen, winsize, gamesize)
	game.mainLoop()

