# Smileys Be Gone!
# A game with the goal to make smileys disappear
# Copyright (C) 2006 Neeraj Kumar and Robert Marianski

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import pygame
from pygame.locals import *
from gamestructs import Node, Edge
from gameutils import HORIZONTAL, VERTICAL, make_rect, tupmap
from pickle import dump, load
from operator import attrgetter, neg, add, sub
from gamegraphics import AnimatedSprite, SpriteRect
import os

# enums representing the different basic types
TRAIN, ROAD, PED, BUILDING = range(4)

# for indexing into globs dict
TRAIN_NODE, ROAD_NODE, PED_NODE, BUILDING_NODE, ROAD_EDGE, ROAD_EDGE_VERT, PED_EDGE, PED_EDGE_VERT, PED_EDGE_ROAD, PED_EDGE_ROAD_VERT, BUILDING_EDGE, BUILDING_EDGE_VERT  = range(12)

w = 15	# size of sidewalk

# offset of map to mapbuilder
mb_offset = (25,20)
# offset used to generate background image
bg_offset = (-45,-45)

# the actual game sizes for each component
sizes = {TRAIN: 2*w, ROAD: 2*w, PED: w, BUILDING: 4*w}

# distances in between nodes
dists = {ROAD: 8*w, PED: 5*w, BUILDING: 8*w}

# the globs used for the map builder images
globs = {TRAIN_NODE:         'mb_train_button_[01].png',
         ROAD_NODE:          'mb_road_node_[01].png',
				 PED_NODE:           'mb_ped_node_[01].png',
				 BUILDING_NODE:      'mb_building_node_[01].png',
         ROAD_EDGE:          'mb_road_edge_[01].png',
         ROAD_EDGE_VERT:     'mb_road_edge_vert_[01].png',
				 PED_EDGE:           'mb_ped_edge_[01].png',
				 PED_EDGE_VERT:      'mb_ped_edge_vert_[01].png',
				 PED_EDGE_ROAD:      'mb_ped_edge_road_[01].png',
				 PED_EDGE_ROAD_VERT: 'mb_ped_edge_road_vert_[01].png',
				 BUILDING_EDGE:      'mb_building_edge_[01].png',
				 BUILDING_EDGE_VERT: 'mb_building_edge_vert_[01].png'}

node_images = {TRAIN:    'train_0_0.png',
               ROAD:     'roadnode.png',
							 PED:      'pednode.png',
							 BUILDING: 'buildingnode_0_0.png'}

edge_images = {TRAIN:    ['trainedge_0.png'],
               ROAD:     ['roadedge_0_0.png', 'roadedge_1_0.png'],
               BUILDING: ['buildingedge_0_0.png', 'buildingedge_0_1.png'],
               PED:      [['pededge_0_0.png', 'pededge_1_0.png'],
                         ['crosswalk_0.png', 'crosswalk_1.png']]}

# create global mouse sprite that everyone can access
ms = SpriteRect(Rect(0,0,1,1))

class MapBuilder(object):
	def __init__(self, offset=mb_offset, nrows=7, ncols=7):
		"""initialize the groups needed to make the map builder display

		the make_map function creates the sprites and adds them to the group"""
		self.offset = offset
		self.offx, self.offy = offset
		self.nrows = nrows
		self.ncols = ncols

		# contains list of rects that are used for edge intersection detection
		self.edgerects = []

		# sprite group to use
		g = pygame.sprite.RenderUpdates
		# building nodes, road nodes, ped nodes, train nodes
		self.bnodes = g()
		self.rnodes = g()
		self.pnodes = g()

		# list of all the edges
		#self.bedges = g()	# don't need any building edges
		self.redges = g()
		self.pedges = g()
		self.tedges = g()
		self.bedges = g()

		# groups used to make collision detection easier
		self.inactive_nodes = g()
		self.inactive_edges = g()
		self.active_nodes = g()
		self.active_edges = g()
		self.nodes = g()
		self.edges = g()
		self.horizontal_edges = g()
		self.vertical_edges = g()

		# train buttons off to the side to generate trains
		self.tbuttons = g()

	def __add(self, groupsToAddTo, objs):
		"""helper function to add sprites to a group and inactive group"""
		for group in groupsToAddTo:
			group.add(objs)

	# these _make* functions are helper functions that
	# create the sprite nodes and
	# add themselves to the appropriate groups
	def _make_road_nodes(self):
		"""create all the road node sprites and add to group"""
		roaddist = dists[ROAD]
		x = self.offx + sizes[ROAD]/2
		y = self.offy + sizes[ROAD]/2
		self.__add([self.rnodes, self.nodes, self.inactive_nodes],
		           [NodeSprite(ROAD, (x+row*roaddist,
		                              y+col*roaddist))
		            for row in xrange(self.nrows)
		            for col in xrange(self.ncols)])

	def _make_ped_edges(self):
		"""create the pedestrian edges and add to group"""
		peddist = dists[PED]
		roaddist = dists[ROAD]
		#TODO add helper function to avoid duplication
		# horizontal edges
		x = self.offx + sizes[ROAD] + sizes[PED]/2 + dists[PED]/2
		y = self.offy + sizes[ROAD] + sizes[PED]/2
		# iterate through the locations of the road nodes and
		# add the group of sidewalk nodes appropriately
		for row in range(self.nrows-1):
			for col in range(self.ncols-1):
				pedx = x+row*roaddist
				pedy = y+col*roaddist
				positions = [(pedx,pedy),
										 (pedx,pedy+peddist)]
				self.__add([self.pedges, self.edges, self.inactive_edges, self.horizontal_edges],
									 [NodeSprite(PED_EDGE, pos) for pos in positions])
				if row == self.nrows-2: continue
				# now add the pedestrian edges that go over the road nodes
				positions = [(pedx+roaddist/2,pedy),
				             (pedx+roaddist/2,pedy+peddist)]
				self.__add([self.pedges, self.edges, self.inactive_edges, self.horizontal_edges],
									 [NodeSprite(PED_EDGE_ROAD, pos) for pos in positions])
		# vertical edges
		x = self.offx + sizes[ROAD] + sizes[PED]/2
		y = self.offy + sizes[ROAD] + sizes[PED]/2 + dists[PED]/2
		for row in range(self.nrows-1):
			for col in range(self.ncols-1):
				pedx = x+row*roaddist
				pedy = y+col*roaddist
				positions = [(pedx,pedy),
										 (pedx+peddist,pedy)]
				self.__add([self.pedges, self.edges, self.inactive_edges, self.vertical_edges],
									 [NodeSprite(PED_EDGE_VERT, pos) for pos in positions])
				# now add the pedestrian edges that go over the road nodes
				if col == self.ncols-2: continue
				positions = [(pedx,pedy+roaddist/2),
										 (pedx+peddist,pedy+roaddist/2)]
				self.__add([self.pedges, self.edges, self.inactive_edges, self.vertical_edges],
									 [NodeSprite(PED_EDGE_ROAD_VERT, pos) for pos in positions])

	def _make_ped_nodes(self):
		"""create all the pedestrian node sprites and add group"""
		peddist = dists[PED]
		roaddist = dists[ROAD]
		x = self.offx + sizes[ROAD] + sizes[PED]/2
		y = self.offy + sizes[ROAD] + sizes[PED]/2
		# iterate through the locations of the road nodes and
		# add the group of sidewalk nodes appropriately
		for row in range(self.nrows-1):
			for col in range(self.ncols-1):
				pedx = x+row*roaddist
				pedy = y+col*roaddist
				positions = [(pedx,pedy),
				             (pedx+peddist,pedy),
										 (pedx,pedy+peddist),
										 (pedx+peddist,pedy+peddist)]
				self.__add([self.pnodes, self.nodes, self.inactive_nodes],
									 [NodeSprite(PED, pos) for pos in positions])

	def _make_building_nodes(self):
		"""create the building node sprites and add to group"""
		blddist = dists[BUILDING]
		x = self.offx + sizes[ROAD] + sizes[PED] + sizes[BUILDING]/2
		y = self.offy + sizes[ROAD] + sizes[PED] + sizes[BUILDING]/2
		# building nodes now have to be added to the other sprite groups
		# because of building edges
		self.__add([self.bnodes, self.nodes, self.inactive_nodes],
		           [NodeSprite(BUILDING, (x+row*blddist, y+col*blddist))
		            for row in xrange(self.nrows-1)
		            for col in xrange(self.ncols-1)])

	def _make_building_edges(self):
		"""create the building edges between building nodes"""
		blddist = dists[BUILDING]
		x = self.offx + sizes[ROAD] + sizes[PED] + sizes[BUILDING]/2 + dists[BUILDING]/2
		y = self.offy + sizes[ROAD] + sizes[PED] + sizes[BUILDING]/2
		self.__add([self.bedges, self.edges, self.inactive_edges, self.horizontal_edges],
		           [NodeSprite(BUILDING_EDGE, (x+blddist*row, y+blddist*col))
		            for col in xrange(self.ncols-1)
		            for row in xrange(self.nrows-1)])
		# vertical edges
		x = self.offx + sizes[ROAD] + sizes[PED] + sizes[BUILDING]/2
		y = self.offy + sizes[ROAD] + sizes[PED] + sizes[BUILDING]/2 + dists[BUILDING]/2
		self.__add([self.bedges, self.edges, self.inactive_edges, self.vertical_edges],
		           [NodeSprite(BUILDING_EDGE_VERT, (x+blddist*row, y+blddist*col))
		            for col in xrange(self.ncols-1)
		            for row in xrange(self.nrows-1)])

	def _make_road_edges(self):
		"""create the road edges and add them to group"""
		roaddist = dists[ROAD]
		#TODO add helper function to avoid duplication
		# horizontal edges
		x = self.offx + sizes[ROAD]/2 + dists[ROAD]/2
		y = self.offy + sizes[ROAD]/2
		self.__add([self.redges, self.edges, self.inactive_edges, self.horizontal_edges],
		           [NodeSprite(ROAD_EDGE, (x+roaddist*row, y+roaddist*col))
		            for col in xrange(self.ncols)
		            for row in xrange(self.nrows-1)])
		# vertical edges
		x = self.offx + sizes[ROAD]/2
		y = self.offy + sizes[ROAD]/2 + dists[ROAD]/2
		self.__add([self.redges, self.edges, self.inactive_edges, self.vertical_edges],
		           [NodeSprite(ROAD_EDGE_VERT, (x+roaddist*row, y+roaddist*col))
		            for col in xrange(self.ncols-1)
		            for row in xrange(self.nrows)])

	def make_map(self):
		"""create a visual representation of edges/nodes on the screen"""
		# create the edges
		self._make_road_edges()
		self._make_ped_edges()
		# create the nodes
		self._make_road_nodes()
		self._make_ped_nodes()
		self._make_building_nodes()
		# create the train buttons off to the side
		self._make_train_buttons()
		# building edges
		self._make_building_edges()

	def _make_train_buttons(self):
		"""create the 'T' buttons next to the left of the road nodes"""
		roaddist = dists[ROAD]
		x = self.offx - sizes[ROAD]/2 + 5
		y = self.offy + sizes[ROAD]/2
		self.__add([self.tbuttons],
		           [NodeSprite(TRAIN_NODE, (x,y+i*roaddist))
		           for i in range(self.nrows)])

	def draw(self, surface):
		# we draw edges before nodes to overlap correctly
		draworder = [ self.redges, self.pedges, self.tedges, self.bedges,
									self.bnodes, self.rnodes, self.pnodes, self.tbuttons]
		# return all the rects that were drawn to pass to display.update
		return sum((group.draw(surface) for group in draworder), [])

	def process_create(self):
		"""left mouse button was pressed, we need to create a new object"""
		collide = pygame.sprite.spritecollide
		# check if a train button was clicked
		for tb in collide(ms, self.tbuttons, 0):
			#TODO horizontal trains only for now?
			trainedge = self._drawTrain(tb.pos[1])
			if trainedge:
				tb.activate()
				# check if adding the edge crosses an existing edge
				for idx in trainedge.rect.collidelistall(self.edgerects):
					otheredge = self.edgerects[idx]
					self.__activateIntersectionNode(trainedge.rect, otheredge)
			return
		# check for adding a node, the trivial case
		for node in collide(ms, self.inactive_nodes, 0):
			node.activate()
			self.inactive_nodes.remove(node)
			self.active_nodes.add(node)
			return
		# now check for adding an edge
		# we need to be more careful when adding edges
		# because we need to verify that there is an edge on both sides
		for edge in collide(ms, self.inactive_edges, 0):
			# check if we have some active nodes surrounding the edge
			# and activate all the edges in between
			edgesToAdd = self.__getEdgesToActivate(edge)
			if not edgesToAdd: return
			# check if an intersecting node exists
			rect = self.__createSurroundingRect(edgesToAdd)
			for idx in rect.collidelistall(self.edgerects):
				otheredge = self.edgerects[idx]
				# check to see that the otheredge rect came from
				# the same edge type
				# this is annoying because we have to collide backwards now
				# to get the original edge that this new surrounding rect
				# collided with
				rectedges = [e.rect for e in self.active_edges]
				for i in otheredge.collidelistall(rectedges):
					# and now this is annoying because we have to iterate through
					# because we can't directly index it (no slice operator)
					e = [e for j,e in enumerate(self.active_edges) if j == i][0]
					# intersections with pedestrian edges are ok
					typea, typeb = map(self.__getTypeOf, (edge, e))
					if PED in (typea, typeb): continue
					#if self.__getTypeOf(edge) != self.__getTypeOf(e):
					if typea != typeb:
						return
				self.__activateIntersectionNode(rect, otheredge)
			# add the new surrounding rect to the list of seen rects
			# for intersection purposes
			self.edgerects.append(rect)
			for edge in edgesToAdd:
				edge.activate()
				self.inactive_edges.remove(edge)
				self.active_edges.add(edge)
			return
		# finally check if a building was clicked
		for building in collide(ms, self.bnodes, 0):
			# no special logic for buildings, either active or not
			if not building.active:
				building.activate()

	def process_delete(self):
		"""Right button was clicked. We may need to remove an object"""
		collide = pygame.sprite.spritecollide
		# first check if a train button was clicked
		for tb in collide(ms, self.tbuttons, 0):
			tb.deactivate()
			self._undrawTrain(tb.pos[1])
			return
		# check to see if we got a click on an active node
		for node in collide(ms, self.active_nodes, 0):
			edgesToRemove = self.__getEdgesToRemoveFromNode(node)
			for edge in edgesToRemove:
				# train edges do not get deactivated
				if self.__isTrainEdge(edge): continue
				edge.deactivate()
				self.active_edges.remove(edge)
				self.inactive_edges.add(edge)
				self.__removeSurroundingEdgeRect(edge)
			node.deactivate()
			self.active_nodes.remove(node)
			self.inactive_nodes.add(node)
			return
		# check if we got a click on an active edge
		for edge in collide(ms, self.active_edges, 0):
			# we now have to check if there are more edges that have to be
			# removed because of this remove
			edgesToRemove = self.__getEdgesToRemove(edge)
			if not edgesToRemove: return
			self.__removeSurroundingEdgeRect(edge)
			for edge in edgesToRemove:
				if self.__isTrainEdge(edge): continue
				edge.deactivate()
				self.active_edges.remove(edge)
				self.inactive_edges.add(edge)
			return
		# finally check if a building was clicked on
		for building in collide(ms, self.bnodes, 0):
			# no special logic for buildings, either active or not
			if building.active:
				building.deactivate()

	def process_mouseclick(self, event):
		"""handle adding/removing nodes/edges"""
		# first check the create events
		if event.button == 1:
			self.process_create()
		# right click means to remove objects
		elif event.button == 3:
			self.process_delete()

	def __activateIntersectionNode(self, rect, otheredge):
		"""activate the node from the intersecting rect and edge"""
		intersection = rect.clip(otheredge)
		intersection = intersection.center
		for node in self.inactive_nodes:
			if node.active: continue	# we want to turn on inactive nodes
			if node.pos == intersection:
				node.activate()
				self.inactive_nodes.remove(node)
				self.active_nodes.add(node)

	def __removeSurroundingEdgeRect(self, edge):
		"""remove the surrounding edge rect from the list"""
		indices = edge.rect.collidelistall(self.edgerects)
		self.edgerects = [e for i,e in enumerate(self.edgerects)
											if i not in indices]

	def __createSurroundingRect(self, edges):
		"""create a rect around all the edges"""
		return edges[0].rect.unionall(edges)

	def __isTrainEdge(self, edge):
		"""returns whether the edge is a train edge"""
		for _ in (e for e in self.tedges if e.pos[1] == edge.pos[1]):
			return 1
		return 0

	def __getOrientationIndex(self, edge):
		"""return the appropriate index into the position rect for edge's orientation"""
		if edge in self.horizontal_edges: return 1
		elif edge in self.vertical_edges: return 0
		else: raise ValueError("Unknown edge: %s" % edge)

	def __getEdgesToRemoveFromNode(self, node):
		"""return the edges (if any) to remove as a result of a node getting removed"""
		return self.__getEdgesToRemoveFromNodeWithOrientation(node, 0) + \
		       self.__getEdgesToRemoveFromNodeWithOrientation(node, 1)

	def __getEdgesToRemoveFromNodeWithOrientation(self, node, i):
		"""helper function to return edges to remove based on orientation i"""
		type = self.__getTypeOf(node)
		edgesOnLevel = [e for e in self.active_edges if e.active and e.pos[i] == node.pos[i] and self.__getTypeOf(e) == type]
		# handle trivial case of no edges on node's level
		if not edgesOnLevel: return []
		nodesOnLevel = [n for n in self.active_nodes if n.active and n.pos[i] == node.pos[i]]
		# use orientation for comparison purposes now
		i = 1-i
		lessNodes = [n for n in nodesOnLevel if n.pos[i] < node.pos[i]]
		moreNodes = [n for n in nodesOnLevel if n.pos[i] > node.pos[i]]
		# we may not have nodes on both sides, so we catch the index error
		# and set the value appropriately (list comp section will be false)
		try: minofmaxs = reduce(lambda x,y: min(x,y.pos[i]), moreNodes, moreNodes[0].pos[i])
		except IndexError: minofmaxs = 0
		try: maxofmins = reduce(lambda x,y: max(x,y.pos[i]), lessNodes, lessNodes[0].pos[i])
		except IndexError: maxofmins = 2000		# big number off screen
		return [e for e in edgesOnLevel
		        if (e.pos[i] > node.pos[i] and e.pos[i] < minofmaxs)
		        or (e.pos[i] < node.pos[i] and e.pos[i] > maxofmins)]

	def __getEdgesToRemove(self, edge):
		"""return a list of all edges that should be removed because of this remove operation"""
		i = self.__getOrientationIndex(edge)
		edgesOnLevel = [e for e in self.edges if e.pos[i] == edge.pos[i] and e.active]
		# handle trivial case with one match
		if len(edgesOnLevel) == 1: return edgesOnLevel
		nodesOnLevel = [n for n in self.nodes if n.pos[i] == edge.pos[i] and n.active]
		i = 1-i		# invert index to start comparisons
		lessNodes = [n for n in nodesOnLevel if n.pos[i] < edge.pos[i]]
		moreNodes = [n for n in nodesOnLevel if n.pos[i] > edge.pos[i]]
		assert lessNodes and moreNodes and "Illegal state: edge does not have nodes on both sides"
		maxofmins = reduce(lambda x,y: max(x,y.pos[i]), lessNodes, lessNodes[0].pos[i])
		minofmaxs = reduce(lambda x,y: min(x,y.pos[i]), moreNodes, moreNodes[0].pos[i])
		return [e for e in edgesOnLevel if maxofmins < e.pos[i] < minofmaxs]

	def __getEdgesToActivate(self, edge):
		"""check to see if there are active nodes on both sides of edge
		
		   returns the list of edges that should be made active """
		# set the orientation appropriately (i == index of position tuple)
		i = self.__getOrientationIndex(edge)
		# first check if there are is a train edge on this level
		# if there is, then we can't add any edges
		# only need to check horizontal direction
		# because we can't have vertical trains
		if i == 1:
			for te in (e for e in self.tedges if e.pos[i] == edge.pos[i]):
				return
		nodesOnLevel = [n for n in self.nodes if n.pos[i] == edge.pos[i] and n.active]
		type = self.__getTypeOf(edge)
		edgesOnLevel = [e for e in self.edges if e.pos[i] == edge.pos[i] and not e.active and self.__getTypeOf(e) == type]
		# we invert the position idx i to verify location
	  # (because at this point all previous i indices are the same)
		i = 1-i
		less = []
		more = []
		less = [n for n in nodesOnLevel if n.pos[i] < edge.pos[i]]
		if not less: return []
		more = [n for n in nodesOnLevel if n.pos[i] > edge.pos[i]]
		if not more: return []
		# at this point we know that there are some valid edges that can
		# be drawn
		# return the edges that are on the same level and are between
		# the surrounding nodes

		maxofmins = reduce(lambda x,y: max(x,y.pos[i]), less, less[0].pos[i])
		minofmaxs = reduce(lambda x,y: min(x,y.pos[i]), more, more[0].pos[i])
		return [e for e in edgesOnLevel if maxofmins < e.pos[i] < minofmaxs]

	#TODO hardcoded game size here
	def _createBackgroundImage(self, nodes, edges, size=(600,600)):
		"""create a background image suitable for use by the game"""
		from bz2 import BZ2File
		background = pygame.image.load(os.path.join("images", "ground.png"))
		nodesToDrawLast = (ROAD, PED)
		def drawnode():
				"""helper function to avoid duplication"""
				imagename = os.path.join('images', node_images[type])
				image = pygame.image.load(imagename)
				rect = n.getRect()
				background.blit(image, rect)
		for type in nodes:
			# road nodes and ped nodes have to be drawn last
			if type in nodesToDrawLast: continue
			for n in nodes[type]:
				drawnode()
		for type in edges:
			for e in edges[type]:
				# need special logic for ped edges crossing roads
				if type == PED:
					imagename = os.path.join('images', edge_images[PED][e.crosswalk][e.orientation])
				else:
					imagename = os.path.join('images',
																	 edge_images[type][e.orientation])
				nedgeimages = int(e.size / sizes[type])
				rect = e.getRect()
				pos = rect.topleft
				# adjust for drawing multiple images by creating an offset
				# moving function
				assert e.orientation in (HORIZONTAL, VERTICAL)
				if e.orientation == HORIZONTAL:
					image_translator = lambda (x,y),idx: (x+idx*sizes[type],y)
				else:
					image_translator = lambda (x,y),idx: (x,y+idx*sizes[type])
				image = pygame.image.load(imagename)
				for i in range(nedgeimages):
					background.blit(image, image_translator(pos,i))
		# draw road nodes again afterwards, because they get overwritten by train edges
		for type in nodesToDrawLast:
			for n in nodes[type]:
					drawnode()
		path = os.path.join('levels', self.levelname + ".bgi")
		f = BZ2File(path, 'w')
		self.__saveImage(background, f)

	def __saveImage(self, background, f):
		"""save an image to a common shared format"""
		s = pygame.image.tostring(background, "RGB")
		dump(s, f, 1)

	def __getEdgeTypeWithOrientation(self, type, orientation):
		"""return the type with an orientation"""
		# orientation == 1 => horizontal
		# orientation == 0 => vertical
		d = {ROAD: [ROAD_EDGE_VERT, ROAD_EDGE],
		     PED:  [PED_EDGE_VERT, PED_EDGE]}
		try:
			return d[type][orientation]
		except KeyError:
			raise ValueError("Invalid type/orientation: %d,%d" % (type, orientation))

	def __getTypeOf(self, sprite):
		"""determine type of a given sprite

		Easier to do this in the map builder class than sprite because
		it contains all the groups, even though it makes more sense to do
		it in the sprite class"""
		groupTypeDict = {self.rnodes: ROAD,
		                 self.redges: ROAD,
		                 self.bnodes: BUILDING,
		                 self.bedges: BUILDING,
		                 self.pnodes: PED,
		                 self.pedges: PED,
		                 self.tedges: TRAIN}
		for g in sprite.groups():
			try: return groupTypeDict[g]
			except KeyError: continue
		raise ValueError("Unknown sprite type: " % sprite)

	def generate(self):
		"""pickle the currently displayed nodes/edges as well as generate the background for the game"""
		nodes = {TRAIN: [], ROAD: [], PED: [], BUILDING: []}
		edges = {TRAIN: [], ROAD: [], PED: [], BUILDING: []}
		for n in self.active_nodes:
			type = self.__getTypeOf(n)
			# take the offset into account
			x,y = n.pos
			x -= self.offx
			y -= self.offy
			gamenode = Node(len(nodes[type]), type, (x,y), sizes[type])
			nodes[type].append(gamenode)
		# edges are more tricky
		# because we have to aggregate adjacent edges
		edges_seen = []
		for e in self.active_edges:
			if e in edges_seen: continue
			type = self.__getTypeOf(e)
			# special logic for train edges
			if type == TRAIN:
				trainNodeDist = 200
				# hardcoded game size here
				screenwidth = 600
				pos = self.__unoffset(e.pos)[1]
				# create the 2 game nodes and game edge in between
				leftnode = Node(len(nodes[type]), type, (-trainNodeDist, pos), sizes[type])
				rightnode = Node(len(nodes[type]), type, (screenwidth+trainNodeDist, pos), sizes[type])
				gameedge = Edge(len(edges[type]), [leftnode, rightnode])
				# now add them all to the appropriate structures
				edges[type].append(gameedge)
				for n in (leftnode, rightnode):
					nodes[type].append(n)
					if gameedge not in n:
						n.addEdge(gameedge)
				continue
			agg_edges = self.__aggregateEdges(e)
			assert agg_edges
			# don't process any other edges in agg_edges
			edges_seen.extend(agg_edges)
			edge_size = len(agg_edges) * dists[type]
			sns = self.__getSurroundingNodes(e)
			assert len(sns) == 2
			# Convert the surrounding sprite nodes into their corresponding
			# game nodes. We don't want to create them again, because we
			# already have them in the nodes dictionary
			gameNodes = [gamenode for gamenode in nodes[type]
			                      for spritenode in sns
			                      if self.__offset(gamenode.pos) == spritenode.pos]
			assert len(gameNodes) == 2
			# create the new game edge
			gameedge = Edge(len(edges[type]), gameNodes)
			if type == PED:
				gameedge.crosswalk = self.__edgeCrossesRoad(e) and 1 or 0
			# add the edge to the surrounding nodes
			for n in gameNodes:
				# only add an edge if the node doesn't already have it
				if gameedge in n: continue
				n.addEdge(gameedge)
			edges[type].append(gameedge)
		# remove the nodes that have no edges associated with them
		for type in nodes:
			# buildings don't need to have edges to exist
			if type in (BUILDING, TRAIN): continue
			nodes[type] = [n for n in nodes[type] if n.edges]
		#print 'pickling nodes:\n%s' % '\n'.join(['%s = %s' % (k,v) for k,v in nodes.items()])
		#print 'pickling edges:\n%s' % '\n'.join(['%s = %s' % (k,v) for k,v in edges.items()])
		fname = os.path.join('levels', self.levelname + ".dat")
		# move nodes/edges an additional offset for background image
		nodes,edges = self.__applyoffset(nodes,edges)
		# since the offset has been applied, mark negative nodes as spawn
		isneg = lambda (x,y):x<0 or y<0
		istoobig = lambda (x,y):x>600 or y>600
		isoffscreen = lambda x:isneg(x) or istoobig(x)
		for lst in nodes.itervalues():
			for n in (n for n in lst if isoffscreen(n.pos)):
				n.isSpawn = 1
		# fix the edge offset of .5 per pixel
		normalize = lambda x:x==int(x) and x or x+.5
		for _,lst in edges.items():
				for e in lst:
					e.pos = tupmap(normalize, e.pos)
		dump((nodes, edges), open(fname, 'wb'), 1)
		self._createBackgroundImage(nodes, edges)

	def __applyoffset(self, nodes, edges, offset=bg_offset):
		"""apply a given offset to the positions of all game nodes and edges"""
		def applyoffset(d):
			newdict = {TRAIN: [], ROAD: [], PED: [], BUILDING: []}
			for type, lst in d.items():
				for x in lst:
					x.pos = tupmap(add, x.pos, offset)
					newdict[type].append(x)
			return newdict
		return map(applyoffset, (nodes, edges))

	def __edgeCrossesRoad(self, e):
		"""returns if PED edge e crosses a road edge"""
		assert self.__getTypeOf(e) == PED
		for ce in pygame.sprite.spritecollide(e, self.redges, 0):
			if ce.active: return 1
		return 0

	def __offset(self, offset):
		"""helper function to translate a position by the given offset"""
		return tupmap(add, offset, self.offset)

	def __unoffset(self, offset):
		"""take an offset position and translate back into real coordinates"""
		return tupmap(sub, offset, self.offset)

	def __aggregateEdges(self, edge):
		"""return a list of all sprite edges that make up a logical edge"""
		i = self.__getOrientationIndex(edge)
		type = self.__getTypeOf(edge)
		edges = [e for e in self.active_edges if e.pos[i] == edge.pos[i] and self.__getTypeOf(e) == type]
		i = 1-i	# compare other position parameter now
		sns = self.__getSurroundingNodes(edge)
		assert len(sns) == 2 and "edge must have 2 active surrounding nodes"
		a = sns[0].pos[i]; b = sns[1].pos[i]
		assert a < b and "surrounding nodes must be in order"
		return [e for e in edges if a < e.pos[i] < b]

	def __getSurroundingNodes(self, edge):
		"""return the nodes surrounding a particular edge"""
		i = self.__getOrientationIndex(edge)
		type = self.__getTypeOf(edge)
		nodesOnLevel = [n for n in self.active_nodes
		                if edge.pos[i] == n.pos[i]
		                and self.__getTypeOf(n) == type]
		assert len(nodesOnLevel) >= 2 and "valid edges must contain at least 2 surrounding nodes"
		i = 1-i	# revert comparisons to other position component on same level
		less = [n for n in nodesOnLevel if n.pos[i] < edge.pos[i]]
		more = [n for n in nodesOnLevel if n.pos[i] > edge.pos[i]]
		assert less and more and "nodes must exist on both sides of edge"
		#pos = lambda x:x.pos[i]
		#cmp = lambda x,y:op(pos(x),pos(y))
		#op = lambda x,y:x<y
		#lessNode = reduce(cmp, less)
		#op = lambda x,y:cmp(x,y) and x or y
		#moreNode = reduce(cmp, more)
		#return [lessNode, moreNode]
		lessNode = less[0]
		for n in less[1:]:
			if n.pos[i] > lessNode.pos[i]:
				lessNode = n
		moreNode = more[0]
		for n in more[1:]:
			if n.pos[i] < moreNode.pos[i]:
				moreNode = n
		return [lessNode, moreNode]

	# trains are currently only horizontal
	def _drawTrain(self, ypos):
		"""attempt to draw a train in the given road row for the ypos"""
		# first check if there is already a train edge on the row
		for _ in (t for t in self.tedges if t.pos[1] == ypos):
			return	# train edge already exists
		# if there are any edges on the current row
		# that means we can't have a train edge here
		for _ in (e for e in self.redges if e.active and e.pos[1] == ypos):
			return None
		# check if there are any building edges on the current ypos
		for _ in (e for e in self.bedges if e.active and e.pos[1] == ypos and self.__getOrientationIndex(e) == 0):
			return None
		x = self.offx + sizes[ROAD]/2
		r = Rect(x, ypos, (self.ncols-1)*dists[ROAD], 5)
		# now offset the rectangle appropriately (we deal with centers)
		top = r.top
		center = r.centery
		r.move_ip(0, -(center-top))
		trainedge = TrainEdge(r)
		self.edgerects.append(r)
		groups = [self.tedges, self.active_edges, self.edges, self.horizontal_edges]
		self.__add(groups, [trainedge])
		return trainedge

	def _undrawTrain(self, ypos):
		"""remove the train that was draw for the given ypos"""
		for e in self.tedges:
			if e.pos[1] == ypos:
				self.__removeSurroundingEdgeRect(e)
				e.kill()
				return

	def loadState(self, nodes, edges, offset=mb_offset, gameoffset=bg_offset):
		"""Take in a dict of nodes/edges to load the state from. These are the dicts that get pickled"""
		# create a mapping of rects for each game node
		# and see if it collides with any sprite nodes
		# if so, that means the sprite node should be active
		spriteNodes = list(self.inactive_nodes)
		# flip the game offset because we need to unapply it here
		gameoffset = tupmap(neg, gameoffset)
		for type in nodes:
			for n in nodes[type]:
				r = make_rect(n.pos, (1,1))
				r.move_ip(offset)
				r.move_ip(gameoffset)
				#TODO collidelistall -> collidelist
				# currently using listall so assert we have nothing strange going on
				# can remove when satisfied
				indices = r.collidelistall(spriteNodes)
				if not indices: continue
				if indices: assert len(indices) == 1
				n = spriteNodes[indices[0]]
				n.activate()
				self.inactive_nodes.remove(n)
				self.active_nodes.add(n)
		# handle train edges differently (they must be drawn across screen)
		for e in edges[TRAIN]:
			for tb in self.tbuttons:
				# take offset into account
				pos = tb.pos[1] - offset[1] - gameoffset[1]
				if e.pos[1] == pos:
					if self._drawTrain(tb.pos[1]):
						# remember to activate the button if the train can be drawn
						tb.activate()
		# now we handle the edges
		spriteNodes = list(self.inactive_edges)
		spriteNodes.extend(self.tedges)
		for type in edges:
			# train edges were already handled above
			if type == TRAIN: continue
			for e in edges[type]:
				# the rect for edges must be constructed differently
				# because the size does matter
				o = e.orientation
				assert o in (HORIZONTAL, VERTICAL)
				size = o == HORIZONTAL and (e.size,1) or (1, e.size)
				r = make_rect(e.pos, size)
				r.move_ip(offset)
				r.move_ip(gameoffset)
				indices = r.collidelistall(spriteNodes)
				if not indices: continue
				# sprite edges that should be added from game edge
				spriteedges = []
				for se in [spriteNodes[i] for i in indices]:
					# orientation must match to be added
					i = self.__getOrientationIndex(se)
					i = 1-i   # convention is opposite gameutils one
					if not i == o: continue
					spriteedges.append(se)
				# and now we can activate them all
				for spriteedge in spriteedges:
					spriteedge.activate()
					self.inactive_edges.remove(spriteedge)
					self.active_edges.add(spriteedge)
				# the reason we saved the sprite edges that were added was so
				# we could create the surrounding rect for them
				# now create the surrounding rect for the sprite edges
				if not spriteedges: continue
				sr = self.__createSurroundingRect(spriteedges)
				self.edgerects.append(sr)

	def levelchanged(self, levelname):
		"""set the level name to use"""
		self.levelname = levelname

class NodeSprite(AnimatedSprite):
	def __init__(self, type, centerpos):
		glob = globs[type]
		AnimatedSprite.__init__(self, fnameGlob=glob, colorkey=(0,0,0), centerpos=centerpos)
		self.active = 0
	def activate(self):
		self.active = 1
		self.setImage(1)
	def deactivate(self):
		self.active = 0
		self.setImage(0)

class TrainEdge(pygame.sprite.Sprite):
	def __init__(self, r):
		"""create a new train edge at the given location"""
		pygame.sprite.Sprite.__init__(self)
		self.image = pygame.Surface(r.size)
		self.image.fill((255,75,0))
		self.rect = r 
		self.pos = r.center
		self.type = TRAIN
		# train edges are always active, or they are not in any groups
		self.active = 1

def initPygame():
	"""helper function to initialize pygame and return screen"""
	pygame.init()
	return pygame.display.set_mode((1024,768))

class LevelSprite(pygame.sprite.Sprite):
	"""reponsible for drawing the level name"""
	def __init__(self, level=1):
		pygame.sprite.Sprite.__init__(self)
		self.level = level
		self.font = pygame.font.Font(None, 36)
		self._drawtext()
		self.rect = self.image.get_rect(centerx=900, centery=350)
	def _drawtext(self):
		"""change the text passed on the current level"""
		self.name = "Level%04d" % self.level
		self.image = self.font.render(self.name, 1, (255, 255, 255))
	def gettext(self):
		"""return the text for the given level"""
		return self.name
	def increment(self, x):
		"""increment the level and update the text appropriately"""
		self.level += x
		# set appropriate boundaries
		if self.level < 1: self.level = 1
		if self.level > 1000: self.level = 1000
		self._drawtext()

class IncSprite(pygame.sprite.Sprite):
	"""display up arrow to increment level"""
	def __init__(self, text, **pos):
		pygame.sprite.Sprite.__init__(self)
		self.font = pygame.font.Font(None, 36)
		self.image = self.font.render(text, 1, (255, 255, 255))
		self.rect = self.image.get_rect(**pos)

def make_viewsprite():
	"""create a sprite whose image is a large 600x600 image of a background"""
	outline = pygame.Surface((600,600))
	outline.fill((150,150,150))
	background = pygame.Surface((596,596))
	background.fill((0,0,0))
	outline.blit(background, (2,2))
	s = pygame.sprite.Sprite()
	s.image = outline
	mapoffset = mb_offset
	backoffset = bg_offset
	# inverse background offset because we apply in other direction
	backoffset = tupmap(neg, backoffset)
	pos = tupmap(add, mapoffset, backoffset)
	s.rect = outline.get_rect(topleft=pos)
	return s

def main(mb=None, screen=None, level=1):
	if mb is None:
		screen = initPygame()
		mb = MapBuilder()
		mb.make_map()
	gensprite = AnimatedSprite(fnameGlob='mb_generate.png',
	                           centerpos=(900,100))
	# draw a rectangle around the view for the background image
	viewsprite = make_viewsprite()
	viewrect = pygame.sprite.RenderUpdates()
	viewrect.add(viewsprite)
	# generate button logic
	generateGroup = pygame.sprite.RenderUpdates()
	generateGroup.add(gensprite)
	# logic to handle changing levels
	level = LevelSprite(level)
	levelgroup = pygame.sprite.RenderUpdates()
	levelgroup.add(level)
	incgroup = pygame.sprite.RenderUpdates()
	decgroup = pygame.sprite.RenderUpdates()
	incsprite = IncSprite("/\\", left=975, centery=350)
	decsprite = IncSprite("\\/", right=823, centery=350)
	incgroup.add(incsprite)
	decgroup.add(decsprite)
	mb.levelchanged(level.gettext())
	while 1:
		for event in pygame.event.get():
			if event.type  == QUIT: return
			if event.type == KEYDOWN:
				if event.key == K_ESCAPE: return
			if event.type == MOUSEBUTTONDOWN:
				# figure out what mouse button was pressed, and delegate accordingly
				ms.rect.topleft = event.pos
				collide = pygame.sprite.spritecollide
				if collide(ms, generateGroup, 0):
					mb.generate()
				elif collide(ms, decgroup, 0):
					level.increment(-1)
					mb.levelchanged(level.gettext())
				elif collide(ms, incgroup, 0):
					level.increment(1)
					mb.levelchanged(level.gettext())
				else:
					mb.process_mouseclick(event)
		screen.fill((0,0,0))  # clear screen
		groups = [viewrect, mb, generateGroup, levelgroup, incgroup, decgroup]
		rects = sum((g.draw(screen) for g in groups), [])
		#rects = mb.draw(screen) + generateGroup.draw(screen) + levelgroup.draw(screen)
		pygame.display.update(rects)
		pygame.time.wait(200)

if __name__ == "__main__":
	from gameutils import print_license
	print_license()
	import sys
	# if a command line argument is specified, then that file is assumed to be a level file, and the mapbuilder creates a new background image for that level and quits
	try:
		levelfile = sys.argv[1]
	except:
		main()
	else:
		screen = initPygame()
		mb = MapBuilder()
		nodes,edges = load(open(levelfile, 'rb'))
		mb.make_map()
		mb.loadState(nodes, edges)
		# parsing from a name like: 'levels/Level0001.bgi'
		try: level = int(levelfile[-8:-4])
		except Exception: level = 1
		main(mb, screen, level)

